"""ECWMF HRES forecast data access."""

from __future__ import annotations

import hashlib
import logging
import pathlib
from contextlib import ExitStack
from datetime import datetime
from typing import TYPE_CHECKING, Any

LOG = logging.getLogger(__name__)

import numpy as np
import pandas as pd
import xarray as xr
from overrides import overrides

import pycontrails
from pycontrails.core import cache, datalib
from pycontrails.core.met import MetDataset, MetVariable
from pycontrails.datalib.ecmwf.common import ECMWFAPI, rad_accumulated_to_average
from pycontrails.datalib.ecmwf.variables import (
    PRESSURE_LEVEL_VARIABLES,
    SURFACE_VARIABLES,
    TOAIncidentSolarRadiation,
    TopNetSolarRadiation,
    TopNetThermalRadiation,
)
from pycontrails.utils.iteration import chunk_list
from pycontrails.utils.temp import temp_file
from pycontrails.utils.types import DatetimeLike

if TYPE_CHECKING:
    from ecmwfapi import ECMWFService


def get_forecast_filename(
    forecast_time: datetime, timestep: datetime, cc: str = "A1", S: str = "D", E: str = "1"
) -> str:
    """Create forecast filename from ECMWF dissemination products.

    The following dissemination filename convention is used for the
    transmission of ECMWF dissemination products:

    ```ccSMMDDHHIImmddhhiiE`` where:

        cc is Dissemination stream name
        S is dissemination data stream indicator
        MMDDHHII is month, day, hour and minute on which the products are based
        mmddhhii is month, day, hour and minute on which the products are valid at
            ddhhii is set to “______” for Seasonal Forecasting System products
            ii is set to 01 for high resolution forecast time step zero, type=fc, step=0
        E is the Experiment Version Number’ (as EXPVER in MARS, normally 1)

    Parameters
    ----------
    forecast_time : datetime
        Forecast time to stage
    timestep : datetime
        Time within forecast
    cc : str, optional
        Dissemination stream name.
        Defaults to "A1"
    S : str, optional
        Dissemination data stream indicator.
        Defaults to "D"
    E : str, optional
        Experiment Version Number.
        Defaults to "1"

    Returns
    -------
    str
        Filename to forecast file

    Raises
    ------
    ValueError
    """

    if forecast_time.hour not in [0, 6, 12, 18]:
        raise ValueError("Forecast time must have hour 0, 6, 12, or 18")

    if timestep < forecast_time:
        raise ValueError("Forecast timestep must be on or after forecast time")

    forecast_time_str = forecast_time.strftime("%m%d%H%M")
    if forecast_time.hour in [6, 18]:
        S = "S"

    timestep_str = timestep.strftime("%m%d%H")
    ii = "00"

    # for some reason "ii" is set to 01 for the first forecast timestep
    if forecast_time == timestep:
        ii = "01"

    return f"{cc}{S}{forecast_time_str}{timestep_str}{ii}1"


class HRES(ECMWFAPI):
    """Class to support HRES data access, download, and organization.

    Requires account with ECMWF and API key.

    API credentials set in local ``~/.ecmwfapirc`` file:

    .. code:: json

        {
            "url": "https://api.ecmwf.int/v1",
            "email": "<email>",
            "key": "<key>"
        }

    Credentials can also be provided directly ``url`` ``key``, and ``email`` keyword args.

    See `ecmwf-api-client <https://github.com/ecmwf/ecmwf-api-client>`_ documentation
    for more information.

    Parameters
    ----------
    time : datalib.TimeInput | None
        The time range for data retrieval, either a single datetime or (start, end) datetime range.
        Input must be a datetime-like or tuple of datetime-like
        (datetime, :class:`pandas.Timestamp`, :class:`numpy.datetime64`)
        specifying the (start, end) of the date range, inclusive.
        If ``forecast_time`` is unspecified, the forecast time will
        be assumed to be the nearest synoptic hour: 00, 06, 12, 18.
        All subsequent times will be downloaded for relative to :attr:`forecast_time`.
        If None, ``paths`` must be defined and all time coordinates will be loaded from files.
    variables : datalib.VariableInput
        Variable name (i.e. "air_temperature", ["air_temperature, relative_humidity"])
        See :attr:`pressure_level_variables` for the list of available variables.
    pressure_levels : datalib.PressureLevelInput, optional
        Pressure levels for data, in hPa (mbar)
        Set to -1 for to download surface level parameters.
        Defaults to -1.
    paths : str | list[str] | pathlib.Path | list[pathlib.Path] | None, optional
        Path to CDS NetCDF files to load manually.
        Can include glob patterns to load specific files.
        Defaults to None, which looks for files in the :attr:`cachestore` or CDS.
    grid : float, optional
        Specify latitude/longitude grid spacing in data.
        Defaults to 0.25.
    stream : str, optional
        "oper" = atmospheric model/HRES, "enfo" = ensemble forecast.
        Defaults to "oper" (HRES),
    field_type : str, optional
        Field type can be e.g. forecast (fc), perturbed forecast (pf),
        control forecast (cf), analysis (an).
        Defaults to "fc".
    forecast_time : DatetimeLike, optional
        Specify forecast run by runtime.
        Defaults to None.
    cachestore : cache.CacheStore | None, optional
        Cache data store for staging data files.
        Defaults to :class:`cache.DiskCacheStore`.
        If None, cache is turned off.
    url : str
        Override `ecmwf-api-client <https://github.com/ecmwf/ecmwf-api-client>`_ url
    key : str
        Override `ecmwf-api-client <https://github.com/ecmwf/ecmwf-api-client>`_ key
    email : str
        Override `ecmwf-api-client <https://github.com/ecmwf/ecmwf-api-client>`_ email

    Notes
    -----
    `MARS key word definitions <https://confluence.ecmwf.int/display/UDOC/Identification+keywords>`_

    - `class <https://apps.ecmwf.int/codes/grib/format/mars/class/>`_:
       in most cases this will be operational data, or "od"
    - `stream <https://apps.ecmwf.int/codes/grib/format/mars/stream/>`_:
      "enfo" = ensemble forecast, "oper" = atmospheric model/HRES
    - `expver <https://confluence.ecmwf.int/pages/viewpage.action?pageId=124752178>`_:
      experimental version, production data is 1 or 2
    - `date <https://confluence.ecmwf.int/pages/viewpage.action?pageId=118817289>`_:
      there are numerous acceptible date formats
    - `time <https://confluence.ecmwf.int/pages/viewpage.action?pageId=118817378>`_:
      forecast base time, always in synoptic time (0,6,12,18 UTC)
    - `type <https://confluence.ecmwf.int/pages/viewpage.action?pageId=127315300>`_:
      forecast (oper), perturbed or control forecast (enfo only), or analysis
    - `levtype <https://confluence.ecmwf.int/pages/viewpage.action?pageId=149335319>`_:
      options include surface, pressure levels, or model levels
    - `levelist <https://confluence.ecmwf.int/pages/viewpage.action?pageId=149335403>`_:
      list of levels in format specified by **levtype** `levelist reference `_
    - `param <https://confluence.ecmwf.int/pages/viewpage.action?pageId=149335858>`_:
      list of variables in catalog number, long name or short name
    - `step <https://confluence.ecmwf.int/pages/viewpage.action?pageId=118820050>`_:
      hourly time steps from base forecast time
    - `number <https://confluence.ecmwf.int/pages/viewpage.action?pageId=149335478>`_:
      for ensemble forecasts, ensemble numbers
    - `format <https://confluence.ecmwf.int/pages/viewpage.action?pageId=116970058>`_:
      specify netcdf instead of default grib, DEPRECATED `format reference `_
    - `grid <https://confluence.ecmwf.int/pages/viewpage.action?pageId=123799065>`_:
      specify model return grid spacing

    Local ``paths`` are loaded using :func:`xarray.open_mfdataset`.
    Pass ``xr_kwargs`` inputs to :meth:`open_metdataset` to customize file loading.


    Examples
    --------
    >>> from datetime import datetime
    >>> from pycontrails import GCPCacheStore
    >>> from pycontrails.datalib.ecmwf import HRES

    >>> # Store data files to local disk (default behavior)
    >>> times = (datetime(2021, 5, 1, 2), datetime(2021, 5, 1, 3))
    >>> hres = HRES(times, variables="air_temperature", pressure_levels=[300, 250])

    >>> # Cache files to google cloud storage
    >>> gcp_cache = GCPCacheStore(bucket="contrails-301217-unit-test", cache_dir="ecmwf")
    >>> hres = HRES(
    ...     times,
    ...     variables="air_temperature",
    ...     pressure_levels=[300, 250],
    ...     cachestore=gcp_cache
    ... )
    """

    __slots__ = ("server", "stream", "field_type", "forecast_time", "url", "key", "email")

    #: stream type, "oper" = atmospheric model/HRES, "enfo" = ensemble forecast.
    stream: str

    #: Field type, forecast ("fc"), perturbed forecast ("pf"),
    #: control forecast ("cf"), analysis ("an").
    field_type: str

    #: Handle to ECMWFService client
    server: ECMWFService

    #: Forecast run time, either specified or assigned by the closest previous forecast run
    forecast_time: datetime

    #: User provided ``ECMWFService`` url
    url: str | None

    #: User provided ``ECMWFService`` key
    key: str | None

    #: User provided ``ECMWFService`` email
    email: str | None

    __marker = object()

    def __init__(
        self,
        time: datalib.TimeInput | None,
        variables: datalib.VariableInput,
        pressure_levels: datalib.PressureLevelInput = -1,
        paths: str | list[str] | pathlib.Path | list[pathlib.Path] | None = None,
        cachepath: str | list[str] | pathlib.Path | list[pathlib.Path] | None = None,
        grid: float = 0.25,
        stream: str = "oper",
        field_type: str = "fc",
        forecast_time: DatetimeLike | None = None,
        cachestore: cache.CacheStore | None = __marker,  # type: ignore[assignment]
        url: str | None = None,
        key: str | None = None,
        email: str | None = None,
    ) -> None:
        try:
            from ecmwfapi import ECMWFService
        except ModuleNotFoundError as e:
            raise ModuleNotFoundError(
                "Some `ecmwf` module dependencies are missing. "
                "Please install all required dependencies using `pip install -e .[ecmwf]`"
            ) from e

        # constants
        # ERA5 now delays creating the server attribute until it is needed to download
        # from CDS. We could do the same here.
        self.url = url
        self.key = key
        self.email = email
        self.server = ECMWFService("mars", url=self.url, key=self.key, email=self.email)
        self.paths = paths
        if cachestore is self.__marker:
            cachestore = cache.DiskCacheStore()
        self.cachestore = cachestore

        if time is None and paths is None:
            raise ValueError("Time input is required when paths is None")

        self.timesteps = datalib.parse_timesteps(time, freq="1H")
        self.pressure_levels = datalib.parse_pressure_levels(
            pressure_levels, self.supported_pressure_levels
        )
        self.variables = datalib.parse_variables(variables, self.supported_variables)

        self.grid = datalib.parse_grid(grid, [0.1, 0.25, 0.5, 1])  # lat/lon degree resolution
        self.stream = stream  # "enfo" = ensemble forecast, "oper" = atmospheric model/HRES
        self.field_type = (
            field_type  # forecast (oper), perturbed or control forecast (enfo only), or analysis
        )

        # set specific forecast time is requested
        if forecast_time is not None:
            forecast_time_pd = pd.to_datetime(forecast_time)
            if forecast_time_pd.hour not in [0, 6, 12, 18]:
                raise ValueError("Forecast hour must be on one of 00, 06, 12, 18")

            self.forecast_time = datalib.round_hour(forecast_time_pd.to_pydatetime(), 6)

        # if no specific forecast is requested, set the forecast time using timesteps
        elif self.timesteps:
            # round first element to the nearest 6 hour time (00, 06, 12, 18 UTC) for forecast_time
            self.forecast_time = datalib.round_hour(self.timesteps[0], 6)

        # when no forecast_time or time input, forecast_time is defined in _open_and_cache

    def __repr__(self) -> str:
        base = super().__repr__()
        return (
            f"{base}\n\tForecast time: {getattr(self, 'forecast_time', '')}\n\tSteps:"
            f" {getattr(self, 'steps', '')}"
        )

    @classmethod
    def create_synoptic_time_ranges(
        self, timesteps: list[pd.Timestamp]
    ) -> list[tuple[pd.Timestamp, pd.Timestamp]]:
        """Create synoptic time bounds encompassing date range.

        Extracts time bounds for synoptic time range ([00:00, 11:59], [12:00, 23:59])
        for a list of input timesteps.

        Parameters
        ----------
        timesteps : list[pd.Timestamp]
            List of timesteps formatted as :class:`pd.Timestamps`.
            Often this it the output from `pd.date_range()`

        Returns
        -------
        list[tuple[pd.Timestamp, pd.Timestamp]]
            List of tuple time bounds that can be used as inputs to :class:`HRES(time=...)`
        """
        time_ranges = np.unique(
            [pd.Timestamp(t.year, t.month, t.day, 12 * (t.hour // 12)) for t in timesteps]
        )

        if len(time_ranges) == 1:
            time_ranges = [(timesteps[0], timesteps[-1])]
        else:
            time_ranges[0] = (
                timesteps[0],
                time_ranges[1] - pd.Timedelta(1, "H"),
            )
            time_ranges[1:-1] = [(t, t + pd.Timedelta(11, "H")) for t in time_ranges[1:-1]]
            time_ranges[-1] = (time_ranges[-1], timesteps[-1])

        return time_ranges

    @property
    def hash(self) -> str:
        """Generate a unique hash for this datasource.

        Returns
        -------
        str
            Unique hash for met instance (sha1)
        """
        hashstr = (
            f"{self.__class__.__name__}{self.timesteps}{self.variable_shortnames}"
            f"{self.pressure_levels}{self.grid}{self.forecast_time}{self.field_type}{self.stream}"
        )
        return hashlib.sha1(bytes(hashstr, "utf-8")).hexdigest()

    @property
    def pressure_level_variables(self) -> list[MetVariable]:
        """ECMWF pressure level parameters.

        Returns
        -------
        list[MetVariable] | None
            List of MetVariable available in datasource
        """
        return PRESSURE_LEVEL_VARIABLES

    @property
    def single_level_variables(self) -> list[MetVariable]:
        """ECMWF surface level parameters.

        Returns
        -------
        list[MetVariable] | None
            List of MetVariable available in datasource
        """
        return SURFACE_VARIABLES

    @property
    def supported_pressure_levels(self) -> list[int]:
        """Get pressure levels available from MARS.

        Returns
        -------
        list[int]
            List of integer pressure level values
        """
        return [
            1000,
            950,
            925,
            900,
            850,
            800,
            700,
            600,
            500,
            400,
            300,
            250,
            200,
            150,
            100,
            70,
            50,
            30,
            20,
            10,
            7,
            5,
            3,
            2,
            1,
            -1,
        ]

    @property
    def step_offset(self) -> int:
        """Difference between :attr:`forecast_time` and first timestep.

        Returns
        -------
        int
            Number of steps to offset in order to retrieve data starting from input time.
            Returns 0 if :attr:`timesteps` is empty when loading from :attr:`paths`.
        """
        if self.timesteps:
            return int((self.timesteps[0] - self.forecast_time).total_seconds() // 3600)

        return 0

    @property
    def steps(self) -> list[int]:
        """Forecast steps from :attr:`forecast_time` corresponding within input :attr:`time`.

        Returns
        -------
        list[int]
            List of forecast steps relative to :attr:`forecast_time`
        """
        return [self.step_offset + i for i in range(len(self.timesteps))]

    def list_from_mars(self) -> str:
        """List metadata on query from MARS.

        Returns
        -------
        str
            Metadata for MARS request.
            Note this is queued the same as data requests.
        """
        request = self.generate_mars_request(self.forecast_time, self.steps, request_type="list")

        # hold downloaded file in named temp file
        with temp_file() as mars_temp_filename:
            LOG.debug(f"Performing MARS request: {request}")
            self.server.execute(request, mars_temp_filename)

            with open(mars_temp_filename, "r") as f:
                txt = f.read()

        return txt

    def generate_mars_request(
        self,
        forecast_time: datetime | None = None,
        steps: list[int] | None = None,
        request_type: str = "retrieve",
        request_format: str = "mars",
    ) -> str | dict[str, Any]:
        """Generate MARS request in MARS request syntax.

        Parameters
        ----------
        forecast_time : :class:`datetime`, optional
            Base datetime for the forecast.
            Defaults to :attr:`forecast_time`.
        steps : list[int], optional
            list of steps.
            Defaults to :attr:`steps`.
        request_type : str, optional
            "retrieve" for download request or "list" for metadata request.
            Defaults to "retrieve".
        request_format : str, optional
            "mars" for MARS string format, or "dict" for dict version.
            Defaults to "mars".

        Returns
        -------
        str | dict[str, Any]
            Returns MARS query string if ``request_format`` is "mars".
            Returns dict query if ``request_format`` is "dict"

        Notes
        -----
        Brief overview of `MARS request syntax
        <https://confluence.ecmwf.int/display/WEBAPI/Brief+MARS+request+syntax>`_
        """

        if forecast_time is None:
            forecast_time = self.forecast_time

        if steps is None:
            steps = self.steps

        # set date/time for file
        _date = forecast_time.strftime("%Y%m%d")
        _time = forecast_time.strftime("%H")

        # make request of mars
        request: dict[str, Any] = {
            "class": "od",  # operational data
            "stream": self.stream,
            "expver": "1",  # production data only
            "date": _date,
            "time": _time,
            "type": self.field_type,
            "param": f"{'/'.join(self.variable_shortnames)}",
            "step": f"{'/'.join([str(s) for s in steps])}",
            "grid": f"{self.grid}/{self.grid}",
        }

        if self.pressure_levels != [-1]:
            request["levtype"] = "pl"
            request["levelist"] = f"{'/'.join([str(pl) for pl in self.pressure_levels])}"
        else:
            request["levtype"] = "sfc"

        if request_format == "dict":
            return request

        levelist = f",\n\tlevelist={request['levelist']}" if self.pressure_levels != [-1] else ""
        return (
            f"{request_type},\n\tclass={request['class']},\n\tstream={request['stream']},"
            f"\n\texpver={request['expver']},\n\tdate={request['date']},"
            f"\n\ttime={request['time']},\n\ttype={request['type']},"
            f"\n\tparam={request['param']},\n\tstep={request['step']},"
            f"\n\tgrid={request['grid']},\n\tlevtype={request['levtype']}{levelist}"
        )

    @overrides
    def create_cachepath(self, t: datetime) -> str:
        if self.cachestore is None:
            raise ValueError("self.cachestore attribute must be defined to create cache path")

        # get forecast_time and step for specific file
        datestr = self.forecast_time.strftime("%Y%m%d-%H")

        # get step relative to forecast forecast_time
        step = self.step_offset + self.timesteps.index(t)

        # single level or pressure level
        if self.pressure_levels == [-1]:
            suffix = f"hressl{self.grid}{self.stream}{self.field_type}"
        else:
            suffix = f"hrespl{self.grid}{self.stream}{self.field_type}"

        # return cache path
        return self.cachestore.path(f"{datestr}-{step}-{suffix}.nc")

    @overrides
    def download_dataset(self, times: list[datetime]) -> None:
        """Download data from data source for input times.

        Parameters
        ----------
        times : list[:class:`datetime`]
            List of datetimes to download and store in cache datastore
        """

        # get step relative to forecast forecast_time
        steps = [self.step_offset + self.timesteps.index(t) for t in times]
        LOG.debug(f"Downloading HRES dataset for base time {self.forecast_time} and steps {steps}")

        # download in sets of 24
        if len(steps) > 24:
            for _steps in chunk_list(steps, 24):
                self._download_file(_steps)
        elif len(steps) > 0:
            self._download_file(steps)

    @overrides
    def cache_dataset(self, dataset: xr.Dataset) -> None:
        if self.cachestore is None:
            LOG.debug("Cache is turned off, skipping")
            return

        with ExitStack() as stack:
            # group by hour and save one dataset for each hour to temp file
            time_group, datasets = zip(*dataset.groupby("time", squeeze=False))

            xarray_temp_filenames = [stack.enter_context(temp_file()) for _ in time_group]
            xr.save_mfdataset(datasets, xarray_temp_filenames)

            # put each hourly file into cache
            self.cachestore.put_multiple(
                xarray_temp_filenames,
                [
                    self.create_cachepath(datetime.utcfromtimestamp(tg.tolist() / 1e9))
                    for tg in time_group
                ],
            )

    @overrides
    def open_metdataset(
        self,
        dataset: xr.Dataset | None = None,
        xr_kwargs: dict[str, Any] | None = None,
        **kwargs: Any,
    ) -> MetDataset:
        xr_kwargs = xr_kwargs or {}

        #  short-circuit dataset or file paths if provided
        if dataset is not None:
            ds = self._preprocess_hres_dataset(dataset)

        # load from local paths
        elif self.paths is not None:
            ds = self._open_and_cache(xr_kwargs)

        # download from MARS
        else:
            if self.cachestore is None:
                raise ValueError("Cachestore is required to download data")

            # confirm files are downloaded from CDS or MARS
            self.download(**xr_kwargs)

            # ensure all files are guaranteed to be available locally here
            # this would download a file from a remote (e.g. GCP) cache
            disk_cachepaths = [self.cachestore.get(f) for f in self._cachepaths]

            # open cache files as xr.Dataset
            ds = self.open_dataset(disk_cachepaths, **xr_kwargs)

            # TODO: corner case
            # If any files are already cached, they will not have the version attached
            if "pycontrails_version" not in ds.attrs:
                ds.attrs["pycontrails_version"] = pycontrails.__version__

        # run the same ECMWF-specific processing on the dataset
        mds = self._process_dataset(ds, **kwargs)

        # convert accumulated radiation values to average instantaneous values
        # set minimum for all values to 0
        # !! Note that HRES accumulates from the *start of the forecast*,
        # so we need to take the diff of each accumulated value
        # the 0th value is set to the 1st value so each time step has a radiation value !!
        dt_accumulation = 60 * 60

        for key in [
            TOAIncidentSolarRadiation.standard_name,
            TopNetSolarRadiation.standard_name,
            TopNetThermalRadiation.standard_name,
        ]:
            if key in mds.data:
                if len(mds.data["time"]) < 2:
                    raise RuntimeError(
                        f"HRES datasets with data variable {key} must have at least two timesteps"
                        f" to calculate the average instantaneous value of {key}"
                    )

                # take the difference between time slices
                dkey_dt = mds.data[key].diff("time")

                # set difference value back to the data model
                mds.data[key] = dkey_dt

                # set the 0th value of the data to the 1st difference value
                # TODO: this assumption may not be universally applicable!
                mds.data[key][dict(time=0)] = dkey_dt[dict(time=0)]

                rad_accumulated_to_average(mds, key, dt_accumulation)

        return mds

    def _open_and_cache(self, xr_kwargs: dict[str, Any]) -> xr.Dataset:
        """Open and cache :class:`xr.Dataset` from :attr:`self.paths`.

        Parameters
        ----------
        xr_kwargs : dict[str, Any]
            Additional kwargs passed directly to :func:`xarray.open_mfdataset`.
            See :meth:`open_metdataset`.

        Returns
        -------
        xr.Dataset
            Dataset opened from local paths.
        """

        if self.paths is None:
            raise ValueError("Attribute `self.paths` must be defined to open and cache")

        # if timesteps are defined and all timesteps are cached already
        # then we can skip loading
        if self.timesteps and self.cachestore and not self.list_timesteps_not_cached(**xr_kwargs):
            LOG.debug("All timesteps already in cache store")
            disk_cachepaths = [self.cachestore.get(f) for f in self._cachepaths]
            return self.open_dataset(disk_cachepaths, **xr_kwargs)

        # set default parameters for loading grib files
        xr_kwargs.setdefault("engine", "cfgrib")
        xr_kwargs.setdefault("combine", "nested")
        xr_kwargs.setdefault("concat_dim", "step")
        xr_kwargs.setdefault("parallel", False)
        ds = self.open_dataset(self.paths, **xr_kwargs)

        # set forecast time if its not already defined
        if not getattr(self, "forecast_time", None):
            self.forecast_time = ds["time"].values.astype("datetime64[s]").tolist()

        # check that forecast_time is correct if defined
        # note the "time" coordinate here is the HRES forecast_time
        elif self.forecast_time != ds["time"].values.astype("datetime64[s]").tolist():
            raise ValueError(
                f"HRES.forecast_time {self.forecast_time} is not the same forecast time listed"
                " in file"
            )

        ds = self._preprocess_hres_dataset(ds)

        # set timesteps if not defined
        # note that "time" is now the actual timestep coordinates
        if not self.timesteps:
            self.timesteps = ds["time"].values.astype("datetime64[s]").tolist()

        self.cache_dataset(ds)

        return ds

    def _download_file(self, steps: list[int]) -> None:
        """Download data file for base datetime and timesteps.

        Overwrites files if they already exists.

        Parameters
        ----------
        steps : list[int]
            Steps to download relative to base date
        """
        request = self.generate_mars_request(self.forecast_time, steps)

        # Open ExitStack to control temp_file context manager
        with ExitStack() as stack:
            # hold downloaded file in named temp file
            mars_temp_grib_filename = stack.enter_context(temp_file())

            # retrieve data from MARS
            LOG.debug(f"Performing MARS request: {request}")
            self.server.execute(request, mars_temp_grib_filename)

            # translate into netcdf from grib
            LOG.debug("Translating file into netcdf")
            mars_temp_nc_filename = stack.enter_context(temp_file())
            ds = stack.enter_context(xr.open_dataset(mars_temp_grib_filename, engine="cfgrib"))

            ##### TODO: do we need to store intermediate netcdf file?
            ds.to_netcdf(path=mars_temp_nc_filename, mode="w")

            # open file, edit, and save for each hourly time step
            ds = stack.enter_context(
                xr.open_dataset(mars_temp_nc_filename, engine=datalib.NETCDF_ENGINE)
            )
            #####

            # run preprocessing before cache
            ds = self._preprocess_hres_dataset(ds)

            self.cache_dataset(ds)

    def _preprocess_hres_dataset(self, ds: xr.Dataset) -> xr.Dataset:
        """Process HRES data before caching.

        Parameters
        ----------
        ds : xr.Dataset
            Loaded :class:`xr.Dataset`

        Returns
        -------
        xr.Dataset
            Processed :class:`xr.Dataset`
        """

        if "pycontrails_version" in ds.attrs:
            LOG.debug("Input dataset processed with pycontrails > 0.29")
            return ds

        # for pressure levels, need to rename "level" field
        if self.pressure_levels != [-1]:
            ds = ds.rename({"isobaricInhPa": "level"})

        # for single level, and singular pressure levels, add the level dimension
        if len(self.pressure_levels) == 1:
            ds = ds.expand_dims({"level": self.pressure_levels})

        # for single time, add the step dimension and assign time coords to step
        if ds["step"].size == 1:
            if "step" not in ds.dims:
                ds = ds.expand_dims({"step": [ds["step"].values]})

            ds = ds.assign_coords({"valid_time": ("step", [ds["valid_time"].values])})

        # rename fields and swap time dimension for step
        ds = ds.rename({"time": "forecast_time"})
        ds = ds.rename({"valid_time": "time"})
        ds = ds.swap_dims({"step": "time"})

        # drop step/number
        ds = ds.drop_vars(["step", "number"], errors="ignore")

        ds.attrs["pycontrails_version"] = pycontrails.__version__
        return ds



"""Parameters for CoCiP uncertainty calculations."""

from __future__ import annotations

import logging
from dataclasses import asdict, dataclass
from typing import Any, ClassVar

import numpy as np
import numpy.typing as npt
from scipy import stats
from scipy.stats.distributions import rv_frozen

from pycontrails.models.cocip import cocip_params
from pycontrails.models.cocip.cocip_params import CocipParams

logger = logging.getLogger(__name__)


class habit_dirichlet(rv_frozen):
    r"""Custom dirichlet distribution for habit weight distribution uncertainty.

    Scale the habit distributions by a dirichlet distribution
    with alpha parameter :math:`\alpha_{i} = 0.5 + \text{C} \text{G}_{i}`
    where :math:`\text{G}_{i}` is the approximate habit weight distributions
    defined in :attr:`CocipParams().habit_distributions`.

    References
    ----------
    - Table 2 in :cite:`schumannEffectiveRadiusIce2011`
    """

    def __init__(self, C: float = 96.0):
        self.C = C

    def rvs(self, *args: Any, **kwds: Any) -> npt.NDArray[np.float32]:
        """Generate sample set of habit distributions.

        Sampled using dirichlet distribution.

        Parameters
        ----------
        *args
            Used to create a number of habit distributions
        **kwds
            Passed through to :func:`scipy.stats.dirichlet.rvs()`

        Returns
        -------
        npt.NDArray[np.float32]
            Sampled habit weight distribution with the same shape
            as :attr:`CocipParams().habit_distributions`
        """
        if args or (kwds.get("size") is not None and kwds["size"] > 1):
            raise ValueError("habit_dirichlet distribution only supports creating one rv at a time")

        default_distributions = cocip_params._habit_distributions()
        alpha_i = 0.5 + self.C * default_distributions

        # In the first distribution, we assume all ice particles are droxtals
        # There is no way to quantify the uncertainty in this assumption
        # Consequently, we leave the first distribution in default_distributions
        # alone, and only perturb the rest
        distr_list = [stats.dirichlet(a) for a in alpha_i[1:]]

        habit_weights = default_distributions.copy()
        for i, distr in enumerate(distr_list, start=1):
            habit_weights[i] = distr.rvs(**kwds)

        return habit_weights


@dataclass
class CocipUncertaintyParams(CocipParams):
    """Model parameters for CoCiP epistemic uncertainty.

    Any uncertainty parameters should end with the suffix `uncertainty`. See `__post_init__`.

    Uncertainty parameters take a `scipy.stats.distributions.rv_frozen` distribution.

    Default distributions have mean at CocipParam default values.

    To retain specific parameters as defaults, set the value of the key to None.

    Examples
    --------
    >>> import scipy.stats
    >>> from pycontrails.models.cocip import CocipUncertaintyParams

    >>> # Override the 'initial_wake_vortex_depth' field from
    >>> # CocipParams with a uniform value in [0.4, 0.6]
    >>> distr = scipy.stats.uniform(loc=0.4, scale=0.2)
    >>> params = CocipUncertaintyParams(seed=123, initial_wake_vortex_depth_uncertainty=distr)
    >>> params.initial_wake_vortex_depth
    0.41076420

    >>> # Once seeded, calling the class again gives a new value
    >>> params = CocipUncertaintyParams(initial_wake_vortex_depth=distr)
    >>> params.initial_wake_vortex_depth
    0.43526372

    >>> # To retain the default value, set the uncertainty to None
    >>> params = CocipUncertaintyParams(rf_lw_enhancement_factor_uncertainty=None)
    >>> params.rf_lw_enhancement_factor
    1.0
    """

    #: The random number generator is attached to the class (as opposed to an instance).
    #: If many instances of :class:`CocipUncertaintyParams` are needed, it is best to seed
    #: the random number generator once initially. Reseeding the generator will also
    #: reset it, giving rise to duplicate random numbers.
    rng: ClassVar[np.random.Generator] = np.random.default_rng(None)

    #: Reseed the random generator defined in ``__post_init__``
    seed: int | None = None

    #: Schumann takes ``wind_shear_enhancement_exponent`` = 0.5 and discusses the case of 0 and 2/3
    #: as possibilities.
    #: With a value of 0, wind shear is not enhanced.
    wind_shear_enhancement_exponent_uncertainty: rv_frozen | None = stats.triang(
        loc=0.0, c=CocipParams.wind_shear_enhancement_exponent, scale=1.0
    )

    #: Schumann takes ``initial_wake_vortex_depth`` = 0.5 and discusses some
    #: uncertainty in this value. This parameter should be non-negative.
    initial_wake_vortex_depth_uncertainty: rv_frozen | None = stats.triang(
        loc=0.3, c=CocipParams.initial_wake_vortex_depth, scale=0.4
    )

    #: Schumann takes a default value of 0.1 and describes it as an "important adjustable parameter"
    #: Currently, `CocipParams` uses a default value of 0.5
    sedimentation_impact_factor_uncertainty: rv_frozen | None = stats.norm(
        loc=CocipParams.sedimentation_impact_factor, scale=0.1
    )

    #: Teoh 2022 (to appear) takes values between 70% decrease and 100% increase.
    #: This coincides with the log normal distribution defined below.
    nvpm_ei_n_enhancement_factor_uncertainty: rv_frozen | None = stats.lognorm(
        s=0.15, scale=1 / stats.lognorm(s=0.15).mean()
    )

    #: Scale shortwave radiative forcing.
    #: Table 2 in :cite:`schumannParametricRadiativeForcing2012`
    #: provides relative RMS error for SW/LW fit to the data generated
    #: by `libRadTran <http://www.libradtran.org/doku.php>`_.
    #: We use the average RMS error across all habit types (pg 1397) as the standard deviation
    #: of a normally distributed scaling factor for SW forcing
    rf_sw_enhancement_factor_uncertainty: rv_frozen | None = stats.norm(
        loc=CocipParams.rf_sw_enhancement_factor, scale=0.106
    )

    #: Scale longwave radiative forcing.
    #: Table 2 in :cite:`schumannParametricRadiativeForcing2012` provides relative error for SW/LW
    #: fit to the data generated by `libRadTran <http://www.libradtran.org/doku.php>`_.
    #: We use the average RMS error across all habit types (pg 1397) as the standard deviation
    #: of a normally distributed scaling factor for LW forcing.
    rf_lw_enhancement_factor_uncertainty: rv_frozen | None = stats.norm(
        loc=CocipParams.rf_lw_enhancement_factor, scale=0.071
    )

    #: Scale the habit distributions by a dirichlet distribution
    #: with alpha parameter :math:`\alpha_{i} = 0.5 + \text{C} \text{G}_{i}`
    #: where :math:`\text{G}_{i}` is the approximate habit weight distributions
    #: defined in :attr:`CocipParams().habit_distributions`.
    #: Higher values of :math:`\text{C}` correspond to higher confidence in initial estimates.
    habit_distributions_uncertainty: rv_frozen | None = habit_dirichlet(C=96.0)

    def __post_init__(self) -> None:
        """Override values of model parameters according to ranges."""
        if self.seed is not None:
            # Reset the class variable `rng`
            logger.info("Reset %s random seed to %s", self.__class__.__name__, self.seed)
            self.__class__.rng = np.random.default_rng(self.seed)

        # Override defaults value on `CocipParams` with random parameters
        for param, value in self.rvs().items():
            setattr(self, param, value)

    @property
    def uncertainty_params(self) -> dict[str, rv_frozen]:
        """Get dictionary of uncertainty parameters.

        Method checks for attributes ending in `"_uncertainty"`.

        Returns
        -------
        dict[str, dict[str, Any]]
            Uncertainty parameters and values
        """
        # handle these differently starting in version 0.27.0
        exclude = {"rhi_adj", "rhi_boost_exponent"}

        out = {}

        param_dict = asdict(self)
        for uncertainty_param, dist in param_dict.items():
            if uncertainty_param.endswith("_uncertainty") and dist is not None:
                param = uncertainty_param.split("_uncertainty")[0]
                if param not in exclude and param not in param_dict:
                    raise AttributeError(
                        f"Parameter {param} corresponding to uncertainty parameter "
                        f"{uncertainty_param} does not exist"
                    )

                if not isinstance(dist, rv_frozen):
                    raise AttributeError(
                        f"Uncertainty parameter '{uncertainty_param}' must be instance of "
                        "'scipy.stats.distributions.rv_frozen'"
                    )

                out[param] = dist

        return out

    def rvs(self, size: None | int = None) -> dict[str, float | npt.NDArray[np.float_]]:
        """Call each distribution's `rvs` method to generate random parameters.

        Seed calls to `rvs` with class variable `rng`.

        Parameters
        ----------
        size : None | int, optional
            If specified, an `array` of values is generated for each uncertainty parameter.

        Returns
        -------
        dict[str, float | npt.NDArray[np.float_]]
            Dictionary of random parameters. Dictionary keys consists of names of parameters in
            `CocipParams` to be overridden by random value.

        Examples
        --------
        >>> from pprint import pprint
        >>> from pycontrails.models.cocip import CocipUncertaintyParams
        >>> params = CocipUncertaintyParams(seed=456)
        >>> pprint(params.rvs())
        {'habit_distributions': array([[0.0000000e+00, 0.0000000e+00, 0.0000000e+00, 0.0000000e+00,
                0.0000000e+00, 0.0000000e+00, 1.0000000e+00, 0.0000000e+00],
               [1.5554131e-02, 2.1363135e-01, 7.7715185e-03, 1.7690966e-02,
                3.1576434e-03, 3.2992734e-06, 7.4009895e-01, 2.0921326e-03],
               [3.8193921e-03, 2.1235342e-01, 3.3554080e-04, 5.2846869e-04,
                3.1945917e-01, 4.8709914e-04, 4.6250960e-01, 5.0730183e-04],
               [5.7327619e-04, 4.7781631e-01, 4.2596990e-03, 6.7235163e-04,
                1.4447135e-01, 3.6184600e-01, 1.0150939e-02, 2.1006212e-04],
               [1.5397545e-02, 4.0522218e-01, 4.2781001e-01, 1.4331797e-01,
                7.1088417e-04, 9.4511814e-04, 3.3900745e-03, 3.2062260e-03],
               [7.9063961e-04, 3.0336906e-03, 7.7571563e-04, 2.0577813e-02,
                9.4205803e-01, 4.3379897e-03, 3.6786550e-03, 2.4747452e-02]],
              dtype=float32),
         'initial_wake_vortex_depth': 0.39805019708566847,
         'nvpm_ei_n_enhancement_factor': 0.9371878437312526,
         'rf_lw_enhancement_factor': 1.1017491252832377,
         'rf_sw_enhancement_factor': 0.99721639115012,
         'sedimentation_impact_factor': 0.5071779847244678,
         'wind_shear_enhancement_exponent': 0.34100931239701004}
        """
        return {
            param: distr.rvs(size=size, random_state=self.rng)
            for param, distr in self.uncertainty_params.items()
        }

    def as_dict(self) -> dict[str, Any]:
        """Convert object to dictionary.

        Wrapper around :meth:`ModelBase.as_dict` that removes
        uncertainty specific parameters.

        Returns
        -------
        dict[str, Any]
            Dictionary version of self.
        """
        obj = super().as_dict()

        # remove seed and _uncertainty attributes
        # these will throw an error in `ModelBase._load_params()`
        keys_to_remove = ["seed"]
        for key in obj:
            if key.endswith("_uncertainty"):
                keys_to_remove.append(key)

        for key in keys_to_remove:
            obj.pop(key, None)

        return obj



"""Test pycontrails/core/models.py"""

from __future__ import annotations

from dataclasses import dataclass
from typing import Any

import numpy as np
import pytest

from pycontrails import Flight, MetDataArray, MetDataset, Model, ModelParams
from pycontrails.core.met_var import AirTemperature, MetVariable, SpecificHumidity


@dataclass
class ModelTestGridParams(ModelParams):
    """Grid test model params."""

    param1: str = "param1"
    param2: int = 1


@dataclass
class ModelTestModelParams(ModelParams):
    """Flight test model."""

    param1: str = "param1"
    param2: int = 1
    met_latitude_buffer: tuple[float, float] = (5, 5)


class ModelTestGrid(Model):
    """Test grid-like model."""

    name = "grid"
    long_name = "my grid model"
    default_params = ModelTestGridParams
    met: MetDataset
    met_required = True
    met_variables = (AirTemperature,)
    source: MetDataset

    def eval(self, source: None = None, **params: Any) -> MetDataArray:
        self.set_source()
        self.update_params(params)
        self.source.data["temp"] = self.met.data["air_temperature"]
        return self.source["temp"]


class ModelTestFlight(Model):
    """Test flight-like model."""

    name = "flight"
    long_name = "my flight model"
    default_params = ModelTestModelParams
    met_variables = (AirTemperature,)
    source: Flight

    def eval(self, source: Flight = None, **params: Any) -> Flight:
        self.set_source(source)
        self.update_params(params)
        self.downselect_met()
        return self.source


class ModelTestPresets(Model):
    """Test grid-like model with params."""

    name = "grid"
    long_name = "my grid model with param sets"
    default_params = ModelTestModelParams

    def eval(self, source: None = None, **params: Any) -> MetDataArray:
        pass


class ModelTestNoDefaultParams(Model):
    """Test grid-like model with no params."""

    name = "grid"
    long_name = "no default params"

    def eval(self, source: None = None, **params: Any) -> MetDataArray:
        pass


class ModelTestNoName(Model):
    """Test grid-like model with no name."""

    myname = "grid"

    def eval(self, source: None = None, **params: Any) -> MetDataArray:
        pass


OtherVar = MetVariable(short_name="o", standard_name="other")


class ModelTestGridMissingRequiredVariables(Model):
    """Test grid-like model missing required variables."""

    name = "grid"
    long_name = "missing required met variables"
    met_variables = [AirTemperature, OtherVar]

    def eval(self, source: None = None) -> MetDataArray:
        pass


class ModelTestFlightMissingRequiredVariables(Model):
    """Test flight-like model missing required variables."""

    name = "flight"
    long_name = "missing required met variables"
    met_variables = [AirTemperature, OtherVar]
    default_params = ModelParams

    def eval(self, source: None = None) -> MetDataArray:
        pass


# ----------
# ModelBase Tests
# ----------


def test_model_needs_names(met_era5_fake: MetDataset) -> None:
    """Model requires name properties defined."""
    with pytest.raises(TypeError, match="abstract methods long_name, name"):
        ModelTestNoName(met_era5_fake)  # type: ignore[abstract]


def test_model_bad_params() -> None:
    """Ensure an error is raised when a model sees an unknown parameter."""
    model = ModelTestFlight(interpolation_method="nearest")
    assert isinstance(model, Model)

    with pytest.raises(KeyError, match="interp_method"):
        ModelTestFlight(interp_method="nearest")


def test_model_copy_source(met_era5_fake: MetDataset, flight_fake: Flight) -> None:
    """Model copy_source."""
    flight_model = ModelTestFlight(met_era5_fake, copy_source=False)
    flight_model.eval(source=flight_fake)
    assert flight_model.source is flight_fake

    flight_model2 = ModelTestFlight(met_era5_fake)
    flight_model2.eval(source=flight_fake)
    assert flight_model2.source is not flight_fake
    assert flight_model.source.data == flight_fake.data


def test_model_type_guards(met_era5_fake: MetDataset, flight_fake: Flight) -> None:
    """Model require_met and require_source_type."""

    model = ModelTestFlight(met=None)
    with pytest.raises(ValueError, match="Meteorology"):
        model.require_met()

    model = ModelTestFlight(met=met_era5_fake)
    met = model.require_met()
    assert met is model.met

    with pytest.raises(ValueError, match="pycontrails.core.flight.Flight"):
        model.require_source_type(Flight)

    model.set_source(flight_fake)
    model.require_source_type(Flight)


# ----------------
# Grid-like Tests
# ----------------


def test_model_grid_params() -> None:
    """Model with parameters defined."""
    params = ModelTestGridParams(param1="test")
    params_dict = params.as_dict()
    assert params_dict["param1"] == "test"
    assert params_dict["param2"] == 1


def test_model_no_default_params(met_era5_fake: MetDataset) -> None:
    """Model with no parameters defined."""
    model = ModelTestNoDefaultParams(met=met_era5_fake)
    default_params = ModelParams().as_dict()
    assert model.params == default_params


def test_model_grid_init(met_era5_fake: MetDataset) -> None:
    """Model __init__."""
    model = ModelTestGrid(met=met_era5_fake)
    assert model.name == "grid"
    assert model.long_name == "my grid model"
    assert "param1" in model.params and model.params["param1"] == "param1"
    assert "param2" in model.params and model.params["param2"] == 1

    # override params via `params` dict and kwargs
    model = ModelTestGrid(met=met_era5_fake, params={"param1": "new_param1"}, param2=2)
    assert model.params["param1"] == "new_param1"
    assert model.params["param2"] == 2

    # update grid params
    model.update_params({"param1": "test2"})
    assert model.params["param1"] == "test2"


def test_model_grid_required_met_variables(met_era5_fake: MetDataset) -> None:
    """Model with required variables."""
    model = ModelTestGrid(met=met_era5_fake)
    assert isinstance(model.met, MetDataset)

    with pytest.raises(KeyError, match="other"):
        ModelTestGridMissingRequiredVariables(met_era5_fake)


def test_model_grid_hash(met_era5_fake: MetDataset) -> None:
    """Model hash."""
    grid_model = ModelTestGrid(met=met_era5_fake)
    assert grid_model.hash == "b0c7e9502e5702950a8f8653e28cc812edf3f302"


def test_model_met_not_copied(met_era5_fake: MetDataset) -> None:
    """Model met is never copied."""
    grid_model = ModelTestGrid(met=met_era5_fake)
    assert grid_model.met is met_era5_fake

    met_era5_fake.data = met_era5_fake.data.rename({"air_temperature": "renamed"})
    assert "air_temperature" not in grid_model.met.data and "renamed" in grid_model.met.data


def test_model_met_grid_eval(met_era5_fake: MetDataset, flight_fake: Flight) -> None:
    """Model met is required for ModelTestGrid."""
    grid_model = ModelTestGrid(met=met_era5_fake)
    out = grid_model.eval()

    # source is based on met, but does not contain met data vars
    assert grid_model.source is not grid_model.met
    assert grid_model.source.coords is not grid_model.met.coords
    for k in grid_model.source.coords:
        np.testing.assert_array_equal(grid_model.source.coords[k], grid_model.met.coords[k])

    assert "temp" in grid_model.source and "air_temperature" not in grid_model.source
    assert np.all(grid_model.source["temp"].values == met_era5_fake.data["air_temperature"].values)
    assert np.all(out.data.values == met_era5_fake.data["air_temperature"].values)

    with pytest.raises(
        ValueError, match="Meteorology is required for this model. Specify with ModelTestGrid"
    ):
        ModelTestGrid()

    # override params on eval
    assert grid_model.params["param1"] == "param1"
    grid_model.eval(param1="override")
    assert grid_model.params["param1"] == "override"


# -----------------
# Flight-like Tests
# -----------------


def test_model_flight_params() -> None:
    """Model params."""
    params = ModelTestModelParams(param1="test")
    params_dict = params.as_dict()
    assert params_dict["param1"] == "test"
    assert params_dict["param2"] == 1

    # default interpolation parameters
    assert "interpolation_method" in params_dict and params_dict["interpolation_method"] == "linear"
    assert (
        "interpolation_bounds_error" in params_dict
        and params_dict["interpolation_bounds_error"] is False
    )
    assert "interpolation_fill_value" in params_dict and np.isnan(
        params_dict["interpolation_fill_value"]
    )

    # default met params
    assert params_dict["downselect_met"] is True
    assert params_dict["met_longitude_buffer"] == (0, 0)
    assert params_dict["met_level_buffer"] == (0, 0)
    assert params_dict["met_time_buffer"] == (np.timedelta64(0, "h"), np.timedelta64(0, "h"))

    # override default params from FlighParams
    assert params_dict["met_latitude_buffer"] == (5, 5)


def test_model_flight_init(met_era5_fake: MetDataset) -> None:
    """Model __init__."""
    model = ModelTestFlight(met_era5_fake)
    assert model.name == "flight"
    assert model.long_name == "my flight model"
    assert "param1" in model.params and model.params["param1"] == "param1"
    assert "param2" in model.params and model.params["param2"] == 1

    # default params get set
    assert (
        "interpolation_method" in model.params and model.params["interpolation_method"] == "linear"
    )

    # override params via `params` dict and kwargs
    model = ModelTestFlight(
        met_era5_fake, params={"param1": "new_param1"}, param2=2, interpolation_method="nearest"
    )
    assert model.params["param1"] == "new_param1"
    assert model.params["param2"] == 2
    assert model.params["interpolation_method"] == "nearest"

    model.update_params({"param1": "test2"})
    assert model.params["param1"] == "test2"


def test_model_flight_required_met_variables(met_era5_fake: MetDataset) -> None:
    """Model required met variables."""
    model = ModelTestFlight(met_era5_fake)
    assert isinstance(model.met, MetDataset)

    with pytest.raises(KeyError, match="other"):
        ModelTestFlightMissingRequiredVariables(met_era5_fake)


def test_model_flight_hash(met_era5_fake: MetDataset) -> None:
    """Ensure pinned hash matches as check for model degradation."""
    flight_model = ModelTestFlight(met_era5_fake)
    assert flight_model.hash == "6c777586ca794efb80a3a4b832f5b04f2371bc67"


def test_model_flight_met_not_copied(met_era5_fake: MetDataset) -> None:
    """Model met never copied."""
    flight_model = ModelTestFlight(met_era5_fake)
    assert flight_model.met is met_era5_fake

    met_era5_fake.data = met_era5_fake.data.rename({"air_temperature": "renamed"})
    assert "air_temperature" not in flight_model.met.data and "renamed" in flight_model.met.data


def test_model_flight_eval(met_era5_fake: MetDataset, flight_fake: Flight) -> None:
    """Model eval."""
    flight_model = ModelTestFlight(met_era5_fake)
    assert np.all(flight_model.eval(source=flight_fake)["latitude"] == flight_fake["latitude"])

    # override params on eval
    assert flight_model.params["param1"] == "param1"
    flight_model.eval(source=flight_fake, param1="override")
    assert flight_model.params["param1"] == "override"


def test_model_flight_downselect_met(met_era5_fake: MetDataset, flight_fake: Flight) -> None:
    """Model downselect_met."""
    # clip flight domain so it is definitely smaller than met
    mask = (flight_fake.level < 300) & (flight_fake.level > 200)
    fl2 = flight_fake.filter(mask)

    flight_model = ModelTestFlight(
        met=met_era5_fake,
        met_longitude_buffer=(15, 15),
        met_latitude_buffer=(5, 10),
        met_level_buffer=(0, 0),
        met_time_buffer=(0, np.timedelta64(1, "h")),
    )
    # When eval is called, the model will downselect met
    _ = flight_model.eval(source=fl2)

    assert flight_model.met is not None
    assert flight_model.met is not met_era5_fake
    assert flight_model.met.data["longitude"].max() == 153
    assert flight_model.met.data["latitude"].max() == 88
    assert flight_model.met.data["level"].max() == 300
    assert flight_model.met.data["time"].max() == np.datetime64("2020-01-01T03:00:00")

    assert flight_model.met.data["longitude"].min() == 61
    assert flight_model.met.data["latitude"].min() == 39
    assert flight_model.met.data["level"].min() == 200
    assert flight_model.met.data["time"].min() == np.datetime64("2020-01-01T01:00:00")

    # don't downselect met if param is False
    flight_model = ModelTestFlight(met_era5_fake, downselect_met=False)
    _ = flight_model.eval(source=fl2)
    assert flight_model.met is met_era5_fake


###########################
# Test `intersect_met_variables`` method
###########################


@dataclass
class VerifyMetParams(ModelParams):
    """Model params to test `verify_met`."""

    verify_met: bool = False


class ModelTestVerifyMet(Model):
    """Model to test `verify_met`."""

    name = "flight"
    long_name = "two variable flight model"
    default_params = VerifyMetParams
    met_variables = [AirTemperature, SpecificHumidity]

    source: Flight | MetDataset

    def eval(self, source: Flight | MetDataset) -> Flight | MetDataset:
        self.set_source(source)
        self.set_source_met()
        return self.source


def test_verify_met_all_vars_on_met(met_era5_fake: MetDataset, flight_fake: Flight) -> None:
    """Test `intersect_met_variables` method on `ModelTestVerifyMet`."""
    model = ModelTestVerifyMet(met=met_era5_fake)
    model._verify_met()
    out = model.eval(source=flight_fake)
    assert isinstance(out, Flight)

    out = model.eval(source=met_era5_fake)
    assert isinstance(out, MetDataset)


@pytest.mark.parametrize("var", ModelTestVerifyMet.met_variables)
def test_verify_met_one_vars_on_met(
    met_era5_fake: MetDataset, flight_fake: Flight, var: MetVariable
) -> None:
    """Test `set_source_met` method on `ModelTestVerifyMet` with single met variable."""
    met = met_era5_fake.copy()
    del met.data[var.standard_name]
    assert var.standard_name not in met
    model = ModelTestVerifyMet(met=met)
    with pytest.raises(KeyError, match=var.standard_name):
        model._verify_met()

    flight_fake[var.standard_name] = np.ones(flight_fake.size)
    out = model.eval(flight_fake)
    assert isinstance(out, Flight)

    out = model.eval(source=met_era5_fake)
    assert isinstance(out, MetDataset)


def test_verify_met_when_met_none(flight_fake: Flight) -> None:
    """Test `intersect_met_variables` method on `ModelTestVerifyMet` with `met=None`."""
    model = ModelTestVerifyMet(met=None)

    with pytest.raises(KeyError):
        model.eval(flight_fake)
    flight_fake["air_temperature"] = np.ones(flight_fake.size)

    with pytest.raises(KeyError, match="specific_humidity"):
        model.eval(flight_fake)
    flight_fake["specific_humidity"] = np.zeros(flight_fake.size)

    out = model.eval(flight_fake)
    assert isinstance(out, Flight)



"""Test flightplan module."""

from __future__ import annotations

from pycontrails.core import flightplan


def test_flightplan_one() -> None:
    atc_plan_str = (
        "(FPL-GEC8145-IN -B77L/H-SDE2E3FGHIJ3J4J5M1RWXYZ/SB1D1"
        "-EGGL1040 -N0474F360 IMVUR1Z IMVUR N63 SAM N19 ADKIK DCT "
        "MOPAT DCT  LIMRI/M083F360 DCT 51N020W 47N030W/M083F380 40N040W "
        "34N045W  28N050W/M083F400 24N055W 19N060W DCT AMTTO DCT ANU DCT "
        "-PBN/A1B1C1D1L1O1S1S2 NAV/RNVD1E2A1 DAT/SVM DOF/140501 REG/DALFA "
        "EET/OKAC0037 ORBB0052 LTAA0159 UKFV0308 UKOV0333 LUUU0344 UKLV0406"
        "EPWW0427 ESAA0521 EKDK0540 ENOR0557 SEL/DFBH OPR/GEC RVR/200) "
        "-E/0740 P/3 R/E S/ J/ A/WHITE BLUE TAIL"
    )

    fp_dict = flightplan.parse_atc_plan(atc_plan_str)

    assert fp_dict["callsign"] == "GEC8145"
    assert fp_dict["flight_rules"] == "I"
    assert fp_dict["type_of_flight"] == "N"
    assert fp_dict["type_of_aircraft"] == "B77L"
    assert fp_dict["wake_category"] == "H"
    assert fp_dict["equipment"] == "SDE2E3FGHIJ3J4J5M1RWXYZ"
    assert fp_dict["transponder"] == "SB1D1"
    assert fp_dict["departure_icao"] == "EGGL"
    assert fp_dict["time"] == "1040"
    assert fp_dict["speed_type"] == "N"
    assert fp_dict["speed"] == "0474"
    assert fp_dict["level_type"] == "F"
    assert fp_dict["level"] == "360"
    assert (
        fp_dict["route"]
        == "IMVUR1Z IMVUR N63 SAM N19 ADKIK DCT MOPAT DCT  "
        "LIMRI/M083F360 DCT 51N020W 47N030W/M083F380 40N040W 34N045W  28N050W/M083F400 "
        "24N055W 19N060W DCT AMTTO DCT ANU DCT"
    )
    assert fp_dict["other_info"] == "E/0740 P/3 R/E S/ J/ A/WHITE BLUE TAIL"

    assert (
        flightplan.to_atc_plan(fp_dict)
        == "(FPL-GEC8145-IN\n"
        "-B77L/H-SDE2E3FGHIJ3J4J5M1RWXYZ/SB1D1\n"
        "-EGGL1040\n"
        "-N0474F360 IMVUR1Z IMVUR N63 SAM N19 ADKIK DCT MOPAT DCT  "
        "LIMRI/M083F360 DCT 51N020W 47N030W/M083F380 40N040W 34N045W  "
        "28N050W/M083F400 24N055W 19N060W DCT AMTTO DCT ANU DCT\n"
        "\n"
        "-E/0740 P/3 R/E S/ J/ A/WHITE BLUE TAIL)"
    )


def test_flightplan_two() -> None:
    fp_str = "(FPL-N12345-IG-SR22/L-S/S-KSEA1414-N0220F090 DCT-PAEN0600-DOF/170428 RMK/DO NOT POST)"

    fp_dict = flightplan.parse_atc_plan(fp_str)

    assert fp_dict["callsign"] == "N12345"
    assert fp_dict["flight_rules"] == "I"
    assert fp_dict["type_of_flight"] == "G"
    assert fp_dict["type_of_aircraft"] == "SR22"
    assert fp_dict["wake_category"] == "L"
    assert fp_dict["equipment"] == "S"
    assert fp_dict["transponder"] == "S"
    assert fp_dict["departure_icao"] == "KSEA"
    assert fp_dict["time"] == "1414"
    assert fp_dict["speed_type"] == "N"
    assert fp_dict["speed"] == "0220"
    assert fp_dict["level_type"] == "F"
    assert fp_dict["level"] == "090"
    assert fp_dict["route"] == "DCT"
    assert fp_dict["destination_icao"] == "PAEN"
    assert fp_dict["duration"] == "0600"
    assert fp_dict["other_info"] == "DOF/170428 RMK/DO NOT POST"

    assert (
        flightplan.to_atc_plan(fp_dict)
        == "(FPL-N12345-IG\n"
        "-SR22/L-S/S\n"
        "-KSEA1414\n"
        "-N0220F090 DCT\n"
        "-PAEN0600\n"
        "-DOF/170428 RMK/DO NOT POST)"
    )



"""Algorithmic Climate Change Functions."""

from __future__ import annotations

from dataclasses import dataclass
from typing import Any, overload

import xarray as xr

import pycontrails
from pycontrails.core.flight import Flight
from pycontrails.core.met import MetDataArray, MetDataset, standardize_variables
from pycontrails.core.met_var import (
    AirTemperature,
    EastwardWind,
    Geopotential,
    NorthwardWind,
    RelativeHumidity,
    SpecificHumidity,
)
from pycontrails.core.models import Model, ModelParams
from pycontrails.core.vector import GeoVectorDataset
from pycontrails.datalib import ecmwf

WideBodyJets = {
    "A332",
    "A333",
    "A338",
    "A339",
    "A342",
    "A343",
    "A345",
    "A356",
    "A359",
    "A388",
    "B762",
    "B763",
    "B764",
    "B772",
    "B773",
    "B778",
    "B779",
    "B788",
    "B789",
}
RegionalJets = {"CRJ1", "CRJ2", "CRJ7", "CRJ8", "CRJ9", "CRJX", "E135", "E145", "E170", "E190"}


@dataclass
class ACCFParams(ModelParams):
    """Default ACCF model parameters.

    See `config-user.yml` definition at
    https://github.com/dlr-pa/climaccf
    """

    lat_bound: tuple[float, float] | None = None
    lon_bound: tuple[float, float] | None = None

    efficacy: bool = True
    efficacy_option: str = "lee_2021"

    accf_v: str = "V1.0"

    ch4_scaling: float = 1.0
    co2_scaling: float = 1.0
    cont_scaling: float = 1.0
    h2o_scaling: float = 1.0
    o3_scaling: float = 1.0

    sep_ri_rw: bool = False

    climate_indicator: str = "ATR"

    horizontal_resolution: float = 0.5

    emission_scenario: str = "pulse"

    time_horizon: int = 20

    pfca: str = "PCFA-ISSR"

    merged: bool = True

    #: RHI Threshold
    issr_rhi_threshold: float = 0.9
    issr_temp_threshold: float = 235

    sac_ei_h2o: float = 1.25
    sac_q: float = 43000000.0
    sac_eta: float = 0.3

    nox_ei: str = "TTV"

    PMO: bool = False


class ACCF(Model):
    """Compute Algorithmic Climate Change Functions (ACCF).

    This class is a wrapper over the DLR / UMadrid library
    `climaccf <https://github.com/dlr-pa/climaccf>`__,
    `DOI: 10.5281/zenodo.6977272 <https://doi.org/10.5281/zenodo.6977272>`__

    Parameters
    ----------
    met : MetDataset
        Dataset containing "air_temperature" and "specific_humidity" variables

    References
    ----------
    - :cite:`dietmullerDlrpaClimaccfDataset2022`
    - :cite:`dietmullerPythonLibraryComputing2022`

    """

    name = "accr"
    long_name = "algorithmic climate change functions"
    met_variables = (
        AirTemperature,
        SpecificHumidity,
        ecmwf.PotentialVorticity,
        Geopotential,
        RelativeHumidity,
        NorthwardWind,
        EastwardWind,
        ecmwf.PotentialVorticity,
    )
    sur_variables = (ecmwf.SurfaceSolarDownwardRadiation, ecmwf.TopNetThermalRadiation)
    default_params = ACCFParams

    short_vars = [v.short_name for v in met_variables + sur_variables]

    ds_met: xr.Dataset | None
    ds_sur: xr.Dataset | None

    def __init__(
        self,
        met: MetDataset,
        surface: MetDataset | None = None,
        params: dict[str, Any] = {},
        **params_kwargs: Any,
    ) -> None:
        try:
            import climaccf  # noqa: F401
        except ModuleNotFoundError as e:
            raise ModuleNotFoundError(
                "Requires the climaccf package which can be installed"
                "using pip install pycontrails[accf]"
            ) from e

        # Normalize ECMWF variables
        met = standardize_variables(met, self.met_variables)

        if surface:
            surface = standardize_variables(surface, self.sur_variables)

        # Surpress warning about humdity scaling because that should be eet
        # using ACCF config variables for this model
        try:
            del met.attrs["history"]
        except KeyError:
            pass

        source = met.attrs["met_source"]
        met.attrs["met_source"] = "not_ecmwf"
        super().__init__(met, params=params, **params_kwargs)
        if self.met:
            self.met.attrs["met_source"] = source

        self._update_accf_config()
        if surface:
            self.surface = surface.copy()

        # This variable won't get used since we are not writing the output
        # anywhere, but the library will complain if it's not defined
        self.path_lib = "./"

        self.ds_met = None
        self.ds_sur = None

    @overload
    def eval(self, source: Flight, **params: Any) -> Flight:
        ...

    @overload
    def eval(self, source: GeoVectorDataset, **params: Any) -> GeoVectorDataset:
        ...

    @overload
    def eval(self, source: MetDataset | None = None, **params: Any) -> MetDataArray:
        ...

    def eval(
        self, source: GeoVectorDataset | Flight | MetDataset | None = None, **params: Any
    ) -> GeoVectorDataset | Flight | MetDataArray:
        """Evaluate accfs along flight trajectory or on meteorology grid.

        Parameters
        ----------
        source : GeoVectorDataset | Flight | MetDataset | None, optional
            Input GeoVectorDataset or Flight.
            If None, evaluates at the :attr:`met` grid points.
        **params : Any
            Overwrite model parameters before eval

        Returns
        -------
        GeoVectorDataset | Flight | MetDataArray
            Returns `np.nan` if interpolating outside meteorology grid.

        Raises
        ------
        NotImplementedError
            Raises if input ``source`` is not supported.
        """

        from climaccf.accf import GeTaCCFs

        self.update_params(params)
        self.set_source(source)

        if isinstance(self.source, GeoVectorDataset):
            self.downselect_met()
            if hasattr(self, "surface"):
                self.surface = self.source.downselect_met(self.surface)

        if isinstance(self.source, MetDataset):
            if self.source["longitude"].size > 1:
                hres = abs(self.source["longitude"].data[1] - self.source["longitude"].data[0])
                self.params["horizontal_resolution"] = float(hres)
            elif self.source["latitude"].size > 1:
                hres = abs(self.source["latitude"].data[1] - self.source["latitude"].data[0])
                self.params["horizontal_resolution"] = float(hres)

        self.set_source_met()
        self._update_accf_config()
        self._generate_weather_store()

        # check aircraft type and set in config if needed
        if self.params["nox_ei"] != "TTV":
            if isinstance(self.source, Flight):
                ac = self.source.attrs["aircraft_type"]
                if ac in WideBodyJets:
                    self.p_settings["ac_type"] = "wide-body"
                elif ac in RegionalJets:
                    self.p_settings["ac_type"] = "regional"
                else:
                    self.p_settings["ac_type"] = "single-aisle"
            else:
                self.p_settings["ac_type"] = "wide-body"

        clim_imp = GeTaCCFs(self)
        clim_imp.get_accfs(**self.p_settings)
        aCCFs, _ = clim_imp.get_xarray()

        # assign ACCF outputs to source
        maCCFs = MetDataset(aCCFs)
        for key, arr in maCCFs.data.items():
            # skip met variables
            if key in self.short_vars:
                continue
            if not isinstance(key, str):
                continue

            if isinstance(self.source, GeoVectorDataset):
                self.source[key] = self.source.intersect_met(maCCFs[key])
            else:
                self.source[key] = (maCCFs.dim_order, arr.data)  # type: ignore

            # Tag output with additional attrs when source is MetDataset
            if isinstance(self.source, MetDataset):
                attrs: dict[str, Any] = {
                    "description": self.long_name,
                    "pycontrails_version": pycontrails.__version__,
                }
                if self.met is not None:
                    attrs["met_source"] = self.met.attrs.get("met_source", "unknown")

                self.source[key].data.attrs.update(attrs)

        return self.source  # type: ignore[return-value]

    def _generate_weather_store(self) -> None:
        from climaccf.weather_store import WeatherStore

        # The library does not call the coordinates by name, it just slices the
        # underlying data array, so we need to put them in the expected order.
        # It also needs variables to have the ECMWF short name
        if isinstance(self.met, MetDataset):
            mt = self.met.data.transpose("time", "level", "latitude", "longitude")
            if mt is None or isinstance(mt, xr.Dataset):
                self.ds_met = mt

        if self.ds_met:
            for var in self.ds_met.data_vars:
                matching_variable = [v for v in self.met_variables if var == v.standard_name]
                if matching_variable:
                    self.ds_met = self.ds_met.rename({var: matching_variable[0].short_name})

        if hasattr(self, "surface"):
            self.ds_sur = self.surface.data.squeeze().transpose("time", "latitude", "longitude")
            for var in self.ds_sur.data_vars:
                matching_variable = [v for v in self.sur_variables if var == v.standard_name]
                if matching_variable:
                    self.ds_sur = self.ds_sur.rename({var: matching_variable[0].short_name})
        else:
            self.ds_sur = None

        ws = WeatherStore(
            self.ds_met, self.ds_sur, ll_resolution=self.p_settings["horizontal_resolution"]
        )
        if self.p_settings["lat_bound"] and self.p_settings["lon_bound"]:
            ws.reduce_domain(
                {
                    "latitude": self.p_settings["lat_bound"],
                    "longitude": self.p_settings["lon_bound"],
                }
            )
        self.ds = ws.get_xarray()
        self.variable_names = ws.variable_names
        self.pre_variable_names = ws.pre_variable_names
        self.coordinate_names = ws.coordinate_names
        self.pre_coordinate_names = ws.pre_coordinate_names
        self.coordinates_bool = ws.coordinates_bool
        self.aCCF_bool = ws.aCCF_bool
        self.axes = ws.axes
        self.var_xr = ws.var_xr

    def _update_accf_config(self) -> None:
        # a good portion of these will get ignored since we are not producing an
        # output file, but the library will complain if they aren't defined
        self.p_settings = {
            "lat_bound": self.params["lat_bound"],
            "lon_bound": self.params["lon_bound"],
            "time_bound": None,
            "horizontal_resolution": self.params["horizontal_resolution"],
            "NOx_aCCF": True,
            "NOx&inverse_EIs": self.params["nox_ei"],
            "output_format": "netCDF",
            "mean": False,
            "std": False,
            "merged": self.params["merged"],
            "aCCF-V": self.params["accf_v"],
            "efficacy": self.params["efficacy"],
            "efficacy-option": self.params["efficacy_option"],
            "emission_scenario": self.params["emission_scenario"],
            "climate_indicator": self.params["climate_indicator"],
            "TimeHorizon": self.params["time_horizon"],
            "ac_type": "wide-body",
            "sep_ri_rw": self.params["sep_ri_rw"],
            "PMO": self.params["PMO"],
            "aCCF-scalingF": {
                "CH4": self.params["ch4_scaling"],
                "CO2": self.params["co2_scaling"],
                "Cont.": self.params["cont_scaling"],
                "H2O": self.params["h2o_scaling"],
                "O3": self.params["o3_scaling"],
            },
            "PCFA": self.params["pfca"],
            "PCFA-ISSR": {
                "rhi_threshold": self.params["issr_rhi_threshold"],
                "temp_threshold": self.params["issr_temp_threshold"],
            },
            "PCFA-SAC": {
                "EI_H2O": self.params["sac_ei_h2o"],
                "Q": self.params["sac_q"],
                "eta": self.params["sac_eta"],
            },
            "Chotspots": False,
            "hotspots_binary": True,
            "color": "Reds",
            "geojson": False,
            "save_path": "./",
            "save_format": "netCDF",
        }



"""Airport data support."""

from __future__ import annotations

import numpy as np
import pandas as pd

from pycontrails.core import cache
from pycontrails.physics import geo, units

#: URL for `Our Airports <https://ourairports.com/>`_ database.
#: Fork of the `ourairports-data repository <https://github.com/davidmegginson/ourairports-data>`_.
OURAIRPORTS_DATABASE_URL: str = (
    "https://github.com/contrailcirrus/ourairports-data/raw/main/airports.csv"
)


def _download_ourairports_csv() -> pd.DataFrame:
    """Download CSV file from fork of ourairports-data github."""
    return pd.read_csv(
        OURAIRPORTS_DATABASE_URL,
        usecols=[
            "type",
            "name",
            "latitude_deg",
            "longitude_deg",
            "elevation_ft",
            "iso_country",
            "iso_region",
            "municipality",
            "scheduled_service",
            "gps_code",
            "iata_code",
        ],
    )


def global_airport_database(
    cachestore: cache.CacheStore | None = None, update_cache: bool = False
) -> pd.DataFrame:
    """
    Load and process global airport database from `Our Airports <https://ourairports.com/>`_.

    The database includes coordinates and metadata for 74867 unique airports.

    Parameters
    ----------
    cachestore : cache.CacheStore | None, optional
        Cache store for airport database.
        Defaults to :class:`cache.DiskCacheStore`.
    update_cache : bool, optional
        Force update to cached airports database.

    Returns
    -------
    pd.DataFrame
        Processed global airport database.

        Global airport database.

    Notes
    -----
    As of 2023 March 30, the global airport database contains:

    .. csv-table::
       :header: "Airport Type", "Number"
       :widths: 70, 30

        "small_airport",    39327
        "heliport",         19039
        "closed",           10107
        "medium_airport",   4753
        "seaplane_base",    1133
        "large_airport",    463
        "balloonport",      45

    References
    ----------
    - :cite:`megginsonOpendataDownloadsOurAirports2023`
    """
    cachestore = cachestore or cache.DiskCacheStore()

    cache_key = "ourairports-data_airports.csv"
    if cachestore.exists(cache_key) and not update_cache:
        airports = pd.read_csv(cachestore.path(cache_key))
    else:
        airports = _download_ourairports_csv()
        airports.to_csv(cachestore.path(cache_key), index=False)

    #: Format dataset by renaming columns & filling nan values
    airports.rename(
        columns={"latitude_deg": "latitude", "longitude_deg": "longitude", "gps_code": "icao_code"},
        inplace=True,
    )
    airports["elevation_ft"].fillna(0, inplace=True)

    # Keep specific airport types used by commercial aviation
    select_airport_types = airports["type"].isin(
        ["large_airport", "medium_airport", "small_airport", "heliport"]
    )

    # Keep airports with valid ICAO codes
    select_icao_codes = (airports["icao_code"].str.len() == 4) & (
        airports["icao_code"].str.isalpha()
    )

    # filter airports
    airports = airports.loc[select_airport_types & select_icao_codes]

    # Format dataset
    airports["elevation_m"] = units.ft_to_m(airports["elevation_ft"].to_numpy())
    airports.sort_values(by=["icao_code"], ascending=True, inplace=True)

    return airports.reset_index(drop=True)


def find_nearest_airport(
    airports: pd.DataFrame,
    longitude: float,
    latitude: float,
    altitude: float,
    *,
    bbox: float = 2.0,
) -> str | None:
    r"""
    Find airport nearest to the waypoints.

    Parameters
    ----------
    airports: pd.DataFrame
        Airport database in the format returned from :func:`global_airport_database`.
    longitude: float
        Waypoint longitude, [:math:`\deg`]
    latitude: float
        Waypoint latitude, [:math:`\deg`]
    altitude: float
        Waypoint altitude, [:math:`m`]
    bbox: float
        Search airports within spatial bounding box of ± `bbox` from the waypoint, [:math:`\deg`]
        Defaults to :math:`2\deg`

    Returns
    -------
    str
        ICAO code of nearest airport.
        Returns None if no airport is found within ``bbox``.

    Notes
    -----
    Function will first search for large airports around the waypoint vicinity.
    If none is found, it will search for medium and small airports
    around the waypoint vicinity.

    The waypoint must be below 10,000 feet to increase the
    probability of identifying the correct airport.
    """
    if altitude > 3000:
        raise ValueError(
            f"Altitude ({altitude} m) is too high (> 3000 m) to identify nearest airport."
        )

    is_near_waypoint = airports["longitude"].between(
        (longitude - bbox), (longitude + bbox)
    ) & airports["latitude"].between((latitude - bbox), (latitude + bbox))

    # Find the nearest airport from largest to smallest airport type
    search_priority = ["large_airport", "medium_airport", "small_airport"]

    for airport_type in search_priority:
        is_airport_type = airports["type"] == airport_type
        nearest_airports = airports.loc[is_near_waypoint & is_airport_type]

        if len(nearest_airports) == 1:
            return nearest_airports["icao_code"].values[0]

        elif len(nearest_airports) > 1:
            distance = distance_to_airports(
                nearest_airports,
                longitude,
                latitude,
                altitude,
            )
            i_nearest = np.argmin(distance)
            return nearest_airports["icao_code"].values[i_nearest]

        else:
            continue

    return None


def distance_to_airports(
    airports: pd.DataFrame,
    longitude: float,
    latitude: float,
    altitude: float,
) -> np.ndarray:
    r"""
    Calculate the 3D distance from the waypoint to the provided airports.

    Parameters
    ----------
    airports : pd.DataFrame
        Airport database in the format returned from :func:`global_airport_database`.
    longitude : float
        Waypoint longitude, [:math:`\deg`]
    latitude : float
        Waypoint latitude, [:math:`\deg`]
    altitude : float
        Waypoint altitude, [:math:`m`]

    Returns
    -------
    np.ndarray
        3D distance from waypoint to airports, [:math:`m`]

    See Also
    --------
    :func:`geo.haversine`
    """
    dist_horizontal = geo.haversine(
        np.full(airports["longitude"].shape, longitude),
        np.full(airports["latitude"].shape, latitude),
        airports["longitude"].to_numpy(),
        airports["latitude"].to_numpy(),
    )
    dist_vertical = altitude - airports["elevation_m"].to_numpy()
    return (dist_horizontal**2 + dist_vertical**2) ** 0.5



"""Pycontrails tests."""

import os
import pathlib

import pycontrails

# Overwrite any CDS env configuration
os.environ["CDSAPI_URL"] = "FAKE"
os.environ["CDSAPI_KEY"] = "FAKE"

# Determine if BADA is available
# Set to BADA_CACHE_DIR environment variable or use default local bada path
default_bada_root = pathlib.Path(*pycontrails.__path__).parents[1] / "bada"
BADA_ROOT = pathlib.Path(os.getenv("BADA_CACHE_DIR", default_bada_root))
BADA3_PATH = BADA_ROOT / "bada3"
BADA4_PATH = BADA_ROOT / "bada4"

BADA_AVAILABLE = BADA3_PATH.exists() and BADA4_PATH.exists()


try:
    import open3d
except ModuleNotFoundError:
    OPEN3D_AVAILABLE = False
else:
    OPEN3D_AVAILABLE = True



"""Cocip / pycocip regression testing.

See README.md for setup instructions
"""

from __future__ import annotations

# set up logging
import logging
import os
import pathlib

import pandas as pd
import support

from pycontrails.models.cocip import cocip as _cocip
from pycontrails.models.cocip import contrail_properties
from pycontrails.physics import constants, geo, thermo, units

# monkey patch deprecated methods
from tests import _deprecated

geo.segment_angle = _deprecated.segment_angle
thermo.c_pm = lambda x: constants.c_pd
contrail_properties.segment_length_ratio = _deprecated.segment_length_ratio
contrail_properties.mean_energy_flux_per_m = _deprecated.mean_energy_flux_per_m
_cocip.CONTINUITY_CONVENTION_PARAM = 0
units.m_to_pl = _deprecated.m_to_pl
units.pl_to_m = _deprecated.pl_to_m

from pycontrails.datalib.ecmwf import ERA5
from pycontrails.models import Cocip

LOG = logging.getLogger("pycontrails")
logging.basicConfig(filename="regression.log", level=logging.INFO)

# setup

flights_filename = "flights.csv"
flight_id = None  # run all
# flight_id = "190101_155_ELY027"  # select single flight
# flight_id = "190101_152_DAL178"
# flight_id = "190101_134_DAL64"
# flight_id = "181231_1057_CKS244"
# flight_id = "181231_1058_UPS214"
# flight_id = "190101_10_UAL84"
# flight_id = "181231_1067_SVA020"
# flight_id = "190101_176_UAL972"

input_path = pathlib.Path(os.path.realpath(__file__)).parent / "inputs"
output_path = pathlib.Path(os.path.realpath(__file__)).parent / "outputs"
bada_path = input_path / "bada"

# TODO: attempt validation with CDS derived met data
time = ("2019-01-01 00:00:00", "2019-01-03 01:00:00")  # (start, end) inclusive time range
pressure_levels = [
    # 1000,
    # 975,
    # 950,
    # 925,
    # 900,
    # 875,
    # 850,
    # 825,
    # 800,
    # 775,
    # 750,
    # 700,
    600,
    550,
    500,
    450,
    400,
    350,
    300,
    250,
    225,
    200,
    175,
    150,
    125,
]  # select pressure levels

# _cache = cache.GCPCacheStore(
#     bucket="contrails-301217-ecmwf-data2",
#     cache_dir="era5",
#     read_only=True,
#     show_progress=True,
# )

# ----
# Load Flights
# ----

# load all flight waypoints
flight_waypoints = pd.read_csv(input_path / flights_filename)

# filter for single flight, if specified
if flight_id is not None:
    flight_waypoints = flight_waypoints[flight_waypoints["Flight ID"] == flight_id]

# create a list of flights from each Flight ID in dataframe
flights = [
    support.create_flight(flight_id, df_flight)
    for (flight_id, df_flight) in flight_waypoints.groupby("Flight ID")
]

# ----
# Load outputs
# ----

LOG.info("Loading output data")

# load pycocip output
pycocip_contrails = pd.read_csv(output_path / "contrails.csv")
pycocip_flights = pd.read_csv(output_path / "flights.csv")

# ----
# Load ERA5
# ----
LOG.info("Opening met")
path_prefix = pathlib.Path("inputs")

met_filepaths = [
    input_path / "ERA5_HRES_20190101_Met.nc",
    input_path / "ERA5_HRES_20190102_Met.nc",
    input_path / "ERA5_HRES_20190103_Met.nc",
]
rad_filepaths = [
    input_path / "ERA5_HRES_20190101_Rad.nc",
    input_path / "ERA5_HRES_20190102_Rad.nc",
    input_path / "ERA5_HRES_20190103_Rad.nc",
]

# define ERA5 sources
era5pl = ERA5(
    time=time, variables=Cocip.met_variables, pressure_levels=pressure_levels, path=met_filepaths
)
era5sl = ERA5(time=time, variables=Cocip.rad_variables, path=rad_filepaths)

# create `MetDataset` from sources
met = era5pl.open_metdataset(xr_kwargs=dict(parallel=False))
rad = era5sl.open_metdataset(xr_kwargs=dict(parallel=False))

# preload cocip met
if flight_id is None:
    LOG.info("Preloading cocip met")
    met, rad = Cocip.process_met_datasets(met, rad)


# ----
# Cocip
# ----
params = {
    "bada3_path": bada_path / "bada3",
    "bada4_path": bada_path / "bada4",
    "process_met": flight_id is not None,
    "downselect_met": flight_id is not None,
    "interpolation_fill_value": 0.0,
}

for fl in flights:
    LOG.info(f'Validating {fl.attrs["flight_id"]}')
    cocip = Cocip(met, rad=rad, params=params)
    cocip.eval(source=fl)

    # get output stats
    pycocip_flight_stats = pycocip_flights[
        pycocip_flights["Flight ID"] == fl.attrs["flight_id"]
    ].iloc[0]

    # get flight, if its included
    try:
        pycocip_flight = pd.read_csv(output_path / "flights" / f'{fl.attrs["flight_id"]}.csv')
    except FileNotFoundError:
        pycocip_flight = None

    # get contrail waypoints
    pycocip_contrail = pycocip_contrails[pycocip_contrails["Flight ID"] == fl.attrs["flight_id"]]

    # validate
    try:
        support.validate_output(cocip, pycocip_flight, pycocip_flight_stats, pycocip_contrail)
    except AssertionError as e:
        LOG.info(f'------- {fl.attrs["flight_id"]} invalid')
        LOG.info(e)
        if flight_id is not None:
            raise (e)



"""
Configuration file for the Sphinx documentation builder.

This file only contains a selection of the most common options. For a full
list see the documentation:
https://www.sphinx-doc.org/en/master/usage/configuration.html
"""

from __future__ import annotations

import datetime

import pycontrails

# -- Project information -----------------------------------------------------

project = "pycontrails"
copyright = f"2021-{datetime.datetime.now().year}, Breakthrough Energy"

author = "Breakthrough Energy"
version = pycontrails.__version__
release = pycontrails.__version__

# -- General configuration ---------------------------------------------------

# parsed files
source_suffix = {
    ".rst": "restructuredtext",
    ".md": "markdown",
}

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.todo",
    "sphinx.ext.intersphinx",
    "sphinx.ext.mathjax",
    "sphinx.ext.napoleon",
    "sphinx.ext.extlinks",
    # https://sphinxcontrib-bibtex.readthedocs.io/en/latest/usage.html
    "sphinxcontrib.bibtex",
    # https://nbsphinx.readthedocs.io/en/0.8.5/
    "nbsphinx",
    # https://sphinx-copybutton.readthedocs.io/en/latest/
    "sphinx_copybutton",
    # Markdown parsing, https://myst-parser.readthedocs.io/en/latest/syntax/optional.html
    "myst_parser",
    # https://github.com/jbms/sphinx-immaterial/issues/38#issuecomment-1055785564
    "IPython.sphinxext.ipython_console_highlighting",
    # https://github.com/wpilibsuite/sphinxext-opengraph
    "sphinxext.opengraph",
]

extlinks = {"issue": ("https://github.com/contrailcirrus/pycontrails/issues/%s", "issue %s")}

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = [
    "README.md",
    "examples/README.md",
    "_build",
    "Thumbs.db",
    ".DS_Store",
    "**.ipynb_checkpoints",
]

# suppress warnings during build
suppress_warnings = ["myst.header"]

# Set up mapping for other projects' docs
intersphinx_mapping = {
    "numpy": ("https://numpy.org/doc/stable/", None),
    "pandas": ("https://pandas.pydata.org/pandas-docs/dev/", None),
    "pyproj": ("https://pyproj4.github.io/pyproj/stable/", None),
    "python": ("https://docs.python.org/3/", None),
    "scipy": ("https://docs.scipy.org/doc/scipy/", None),
    "skimage": ("https://scikit-image.org/docs/stable/", None),
    "sklearn": ("https://scikit-learn.org/stable/", None),
    "xarray": ("https://docs.xarray.dev/en/stable/", None),
    "zarr": ("https://zarr.readthedocs.io/en/stable/", None),
    "traffic": ("https://traffic-viz.github.io/", None),
}

# Display todos by setting to True
todo_include_todos = True

# Napoleon configuration - https://www.sphinx-doc.org/en/master/usage/extensions/napoleon.html
napoleon_google_docstring = False
napoleon_numpy_docstring = True
napoleon_use_param = False
napoleon_use_rtype = False
napoleon_preprocess_types = True

# if convenient, set mapping that maps docstring words to
# richer text in sphinx output
# Note the ~ prefix removes the module name in presentation
napoleon_type_aliases = {
    # general terms
    "sequence": ":term:`sequence`",
    "iterable": ":term:`iterable`",
    "callable": ":py:func:`callable`",
    "dict_like": ":term:`dict-like <mapping>`",
    "dict-like": ":term:`dict-like <mapping>`",
    "path-like": ":term:`path-like <path-like object>`",
    "mapping": ":term:`mapping`",
    # pycontrails
    "VectorDataset": "~pycontrails.VectorDataset",
    "GeoVectorDataset": "~pycontrails.GeoVectorDataset",
    "Flight": "~pycontrails.Flight",
    "Aircraft": "~pycontrails.Aircraft",
    "Fuel": "~pycontrails.Fuel",
    "JetA": "~pycontrails.JetA",
    "GCPCacheStore": "~pycontrails.GCPCacheStore",
    "DiskCacheStore": "~pycontrails.DiskCacheStore",
    "MetDataset": "~pycontrails.MetDataset",
    "MetDataArray": "~pycontrails.MetDataArray",
    "MetDataSource": "~pycontrails.MetDataSource",
    "Model": "~pycontrails.Model",
    "ModelParams": "~pycontrails.ModelParams",
    # pycontrails.models
    "BADA3": "~pycontrails.ext.bada.BADA3",
    "BADA4": "~pycontrails.ext.bada.BADA4",
    "BADAFlight": "~pycontrails.ext.bada.BADAFlight",
    "BADAFlightParams": "~pycontrails.ext.bada.BADAFlightParams",
    "BADAGrid": "~pycontrails.ext.bada.BADAGrid",
    "BADAGridParams": "~pycontrails.ext.bada.BADAGridParams",
    "Cocip": "~pycontrails.models.cocip.Cocip",
    "CocipParams": "~pycontrails.models.cocip.CocipParams",
    "CocipUncertaintyParams": "~pycontrails.models.cocip.CocipUncertaintyParams",
    "CocipFlightParams": "~pycontrails.models.cocip.CocipFlightParams",
    "CocipGrid": "~pycontrails.models.cocip.CocipGrid",
    "CocipGridParams": "~pycontrails.models.cocip.CocipGridParams",
    "Emissions": "~pycontrails.models.emissions.Emissions",
    "ISSR": "~pycontrails.models.issr.ISSR",
    "ISSRParams": "~pycontrails.models.issr.ISSRParams",
    "PCR": "~pycontrails.models.pcr.PCR",
    "PCRParams": "~pycontrails.models.pcr.PCRParams",
    "SAC": "~pycontrails.models.sac.SAC",
    "SACParams": "~pycontrails.models.sac.SACParams",
    "PCC": "~pycontrails.models.pcc.PCC",
    "PCCParams": "~pycontrails.models.pcc.PCCParams",
    # pycontrails.utils
    "ArrayScalarLike": "~pycontrails.utils.types.ArrayScalarLike",
    "ArrayLike": "~pycontrails.utils.types.ArrayLike",
    # numpy
    "np.ndarray": "numpy.ndarray",
    "np.datetime64": "numpy.datetime64",
    "np.timedelta64": "numpy.timedelta64",
    # xarray
    "xr.DataArray": "xarray.DataArray",
    "xr.Dataset": "xarray.Dataset",
    # pandas
    "pd.DataFrame": "pandas.DataFrame",
    "pd.Series": "pandas.Series",
    "pd.Timestamp": "pandas.Timestamp",
}


# generate autosummary files into the :toctree: directory
#   See http://www.sphinx-doc.org/en/master/ext/autosummary.html
autosummary_generate = True
autodoc_typehints = "none"

# autodoc options
autoclass_content = "class"  # only include docstring from Class (not __init__ method)
autodoc_inherit_docstrings = False
autodoc_default_options = {
    "members": None,  # means yes/true/on
    "undoc-members": None,
    "show-inheritance": None,
}

# Add references in bibtex format here
# use with :cite:`perez2011python` etc
# See the References section of the README for instructions to add new references
bibtex_bibfiles = ["_static/pycontrails.bib"]
bibtex_reference_style = "author_year"
bibtex_default_style = "unsrt"
bibtex_cite_id = "cite-{key}"

nbsphinx_timeout = 600
nbsphinx_execute = "never"

# Allow headers to be linkable to level 3.
myst_heading_anchors = 3

# Disable myst translations
myst_disable_syntax: list[str] = []

# Optional MyST Syntaxes
myst_enable_extensions: list[str] = []

# https://www.sphinx-doc.org/en/master/usage/configuration.html#confval-pygments_style
# https://pygments.org/styles/
# pygments_style = "monokai"

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.
# See the documentation for a list of builtin themes.
# https://github.com/pradyunsg/furo
html_theme = "furo"
html_title = f"{project} v{release}"

# Adds a "last updated on" to the bottom of each page
# Set to None to disable
html_last_updated_fmt = ""

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
html_theme_options = {
    "source_repository": "https://github.com/contrailcirrus/pycontrails",
    "source_branch": "main",
    "source_directory": "docs/",
    # "sidebar_hide_name": False,   # default
    # "top_of_page_button": "edit", # default
    # this adds a github icon to the footer. Ugly, but useful.
    # See https://pradyunsg.me/furo/customisation/footer/#using-embedded-svgs
    "footer_icons": [
        {
            "name": "GitHub",
            "url": "https://github.com/contrailcirrus/pycontrails",
            "html": """
                <svg stroke="currentColor" fill="currentColor" stroke-width="0" viewBox="0 0 16 16">
                    <path fill-rule="evenodd" d="M8 0C3.58 0 0 3.58 0 8c0 3.54 2.29 6.53 5.47 7.59.4.07.55-.17.55-.38 0-.19-.01-.82-.01-1.49-2.01.37-2.53-.49-2.69-.94-.09-.23-.48-.94-.82-1.13-.28-.15-.68-.52-.01-.53.63-.01 1.08.58 1.23.82.72 1.21 1.87.87 2.33.66.07-.52.28-.87.51-1.07-1.78-.2-3.64-.89-3.64-3.95 0-.87.31-1.59.82-2.15-.08-.2-.36-1.02.08-2.12 0 0 .67-.21 2.2.82.64-.18 1.32-.27 2-.27.68 0 1.36.09 2 .27 1.53-1.04 2.2-.82 2.2-.82.44 1.1.16 1.92.08 2.12.51.56.82 1.27.82 2.15 0 3.07-1.87 3.75-3.65 3.95.29.25.54.73.54 1.48 0 1.07-.01 1.93-.01 2.2 0 .21.15.46.55.38A8.013 8.013 0 0 0 16 8c0-4.42-3.58-8-8-8z"></path>
                </svg>
            """,
            "class": "",
        },
    ],
}


# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
# html_logo = "_static/img/logo.jpg"
html_favicon = "_static/img/favicon.png"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]
html_css_files = ["css/style.css"]
# html_js_files = []

html_sourcelink_suffix = ""



"""Wave-vortex downwash functions."""

from __future__ import annotations

import numpy as np
import numpy.typing as npt

from pycontrails.models.cocip import wind_shear
from pycontrails.physics import constants, thermo


def max_downward_displacement(
    wingspan: npt.NDArray[np.float_] | float,
    true_airspeed: npt.NDArray[np.float_],
    aircraft_mass: npt.NDArray[np.float_] | float,
    air_temperature: npt.NDArray[np.float_],
    dT_dz: npt.NDArray[np.float_],
    ds_dz: npt.NDArray[np.float_],
    air_pressure: npt.NDArray[np.float_],
    effective_vertical_resolution: float,
    wind_shear_enhancement_exponent: float | npt.NDArray[np.float_],
) -> npt.NDArray[np.float_]:
    """
    Calculate the maximum contrail downward displacement after the wake vortex phase.

    Parameters
    ----------
    wingspan : npt.NDArray[np.float_] | float
        aircraft wingspan, [:math:`m`]
    true_airspeed : npt.NDArray[np.float_]
        true airspeed for each waypoint, [:math:`m s^{-1}`]
    aircraft_mass : npt.NDArray[np.float_] | float
        aircraft mass for each waypoint, [:math:`kg`]
    air_temperature : npt.NDArray[np.float_]
        ambient temperature for each waypoint, [:math:`K`]
    dT_dz : npt.NDArray[np.float_]
        potential temperature gradient, [:math:`K m^{-1}`]
    ds_dz : npt.NDArray[np.float_]
        Difference in wind speed over dz in the atmosphere, [:math:`m s^{-1} / m`]
    air_pressure : npt.NDArray[np.float_]
        pressure altitude at each waypoint, [:math:`Pa`]
    effective_vertical_resolution, wind_shear_enhancement_exponent: float
        Passed through to :func:`wind_shear.wind_shear_enhancement_factor`

    Returns
    -------
    npt.NDArray[np.float_]
        Max contrail downward displacement after the wake vortex phase, [:math:`m`]

    References
    ----------
    - :cite:`holzapfelProbabilisticTwoPhaseWake2003`
    - :cite:`schumannContrailCirrusPrediction2012`
    """
    wingspan_arr = np.broadcast_to(wingspan, true_airspeed.shape)
    aircraft_mass_arr = np.broadcast_to(aircraft_mass, true_airspeed.shape)

    rho_air = thermo.rho_d(air_temperature, air_pressure)
    n_bv = thermo.brunt_vaisala_frequency(air_pressure, air_temperature, dT_dz)
    t_0 = get_effective_time_scale(wingspan, true_airspeed, aircraft_mass_arr, rho_air)

    dz_max_strong = get_downward_displacement_strongly_stratified(
        wingspan, true_airspeed, aircraft_mass_arr, rho_air, n_bv
    )

    is_weakly_stratified = n_bv * t_0 < 0.8
    dz_max_weak = get_downward_displacement_weakly_stratified(
        wingspan=wingspan_arr[is_weakly_stratified],
        true_airspeed=true_airspeed[is_weakly_stratified],
        aircraft_mass=aircraft_mass_arr[is_weakly_stratified],
        rho_air=rho_air[is_weakly_stratified],
        n_bv=n_bv[is_weakly_stratified],
        dz_max_strong=dz_max_strong[is_weakly_stratified],
        ds_dz=ds_dz[is_weakly_stratified],
        t_0=t_0[is_weakly_stratified],
        effective_vertical_resolution=effective_vertical_resolution,
        wind_shear_enhancement_exponent=wind_shear_enhancement_exponent,
    )

    dz_max_strong[is_weakly_stratified] = dz_max_weak
    return dz_max_strong


def get_effective_time_scale(
    wingspan: npt.NDArray[np.float_] | float,
    true_airspeed: npt.NDArray[np.float_],
    aircraft_mass: npt.NDArray[np.float_] | float,
    rho_air: npt.NDArray[np.float_],
) -> npt.NDArray[np.float_]:
    r"""
    Calculate the effective time scale of the wake vortex.

    Parameters
    ----------
    wingspan : npt.NDArray[np.float_]
        aircraft wingspan, [:math:`m`]
    true_airspeed : npt.NDArray[np.float_]
        true airspeed for each waypoint, [:math:`m \ s^{-1}`]
    aircraft_mass : npt.NDArray[np.float_]
        aircraft mass for each waypoint, [:math:`kg`]
    rho_air : npt.NDArray[np.float_]
        density of air for each waypoint, [:math:`kg \ m^{-3}`]

    Returns
    -------
    npt.NDArray[np.float_]
        Wake vortex effective time scale

    Notes
    -----
    See section 2.5 (pg 547) of :cite:`schumannContrailCirrusPrediction2012`.

    References
    ----------
    - :cite:`schumannContrailCirrusPrediction2012`
    """
    c = np.pi**4 / 32
    return c * wingspan**3 * rho_air * true_airspeed / (aircraft_mass * constants.g)


def get_downward_displacement_strongly_stratified(
    wingspan: npt.NDArray[np.float_] | float,
    true_airspeed: npt.NDArray[np.float_],
    aircraft_mass: npt.NDArray[np.float_] | float,
    rho_air: npt.NDArray[np.float_],
    n_bv: npt.NDArray[np.float_],
) -> npt.NDArray[np.float_]:
    """
    Calculate the maximum contrail downward displacement under strongly stratified conditions.

    Parameters
    ----------
    wingspan : npt.NDArray[np.float_] | float
        aircraft wingspan, [:math:`m`]
    true_airspeed : npt.NDArray[np.float_]
        true airspeed for each waypoint, [:math:`m s^{-1}`]
    aircraft_mass : npt.NDArray[np.float_] | float
        aircraft mass for each waypoint, [:math:`kg`]
    rho_air : npt.NDArray[np.float_]
        density of air for each waypoint, [:math:`kg m^{-3}`]
    n_bv : npt.NDArray[np.float_]
        Brunt-Vaisaila frequency, [:math:`s^{-1}`]

    Returns
    -------
    npt.NDArray[np.float_]
        Maximum contrail downward displacement, strongly stratified conditions.

    Notes
    -----
    See section 2.5 (pg 547 - 548) of :cite:`schumannContrailCirrusPrediction2012`.

    References
    ----------
    - :cite:`schumannContrailCirrusPrediction2012`
    """
    c = (1.49 * 16) / (2 * np.pi**3)  # This is W2 in Schumann's Fortran code
    return (c * aircraft_mass * constants.g) / (wingspan**2 * rho_air * true_airspeed * n_bv)


def get_downward_displacement_weakly_stratified(
    wingspan: npt.NDArray[np.float_] | float,
    true_airspeed: npt.NDArray[np.float_],
    aircraft_mass: npt.NDArray[np.float_] | float,
    rho_air: npt.NDArray[np.float_],
    n_bv: npt.NDArray[np.float_],
    dz_max_strong: npt.NDArray[np.float_],
    ds_dz: npt.NDArray[np.float_],
    t_0: npt.NDArray[np.float_],
    effective_vertical_resolution: float,
    wind_shear_enhancement_exponent: npt.NDArray[np.float_] | float,
) -> npt.NDArray[np.float_]:
    """
    Calculate the maximum contrail downward displacement under weakly/stably stratified conditions.

    Parameters
    ----------
    wingspan : npt.NDArray[np.float_] | float
        aircraft wingspan, [:math:`m`]
    true_airspeed : npt.NDArray[np.float_]
        true airspeed for each waypoint, [:math:`m s^{-1}`]
    aircraft_mass : npt.NDArray[np.float_] | float
        aircraft mass for each waypoint, [:math:`kg`]
    rho_air : npt.NDArray[np.float_]
        density of air for each waypoint, [:math:`kg m^{-3}`]
    n_bv : npt.NDArray[np.float_]
        Brunt-Vaisaila frequency, [:math:`s^{-1}`]
    dz_max_strong : npt.NDArray[np.float_]
        Max contrail downward displacement under strongly stratified conditions, [:math:`m`]
    ds_dz : npt.NDArray[np.float_]
        Difference in wind speed over dz in the atmosphere, [:math:`m s^{-1} / m`]
    t_0 : npt.NDArray[np.float_]
        Wake vortex effective time scale
    effective_vertical_resolution: float
        Passed through to :func:`wind_shear.wind_shear_enhancement_factor`
    wind_shear_enhancement_exponent: npt.NDArray[np.float_] | float
        Passed through to :func:`wind_shear.wind_shear_enhancement_factor`

    Returns
    -------
    npt.NDArray[np.float_]
        Maximum contrail downward displacement, weakly/stably stratified conditions.

    Notes
    -----
    See section 2.5 (pg 548) of :cite:`schumannContrailCirrusPrediction2012`.

    References
    ----------
    - :cite:`schumannContrailCirrusPrediction2012`
    """
    b_0 = get_wake_vortex_separation(wingspan)
    dz_max = np.maximum(dz_max_strong, 10.0)
    shear_enhancement_factor = wind_shear.wind_shear_enhancement_factor(
        dz_max, effective_vertical_resolution, wind_shear_enhancement_exponent
    )

    # Calculate epsilon and epsilon star
    # In Schumann's Fortran code, epsn = EDR and epsn_st = EPSN
    epsn = turbulent_kinetic_energy_dissipation_rate(ds_dz, shear_enhancement_factor)
    epsn_st = normalized_dissipation_rate(epsn, wingspan, true_airspeed, aircraft_mass, rho_air)
    return b_0 * (7.68 * (1 - 4.07 * epsn_st + 5.67 * epsn_st**2) * (0.79 - n_bv * t_0) + 1.88)


def get_wake_vortex_separation(
    wingspan: npt.NDArray[np.float_] | float,
) -> npt.NDArray[np.float_] | float:
    """
    Calculate the wake vortex separation.

    Parameters
    ----------
    wingspan : npt.NDArray[np.float_] | float
        aircraft wingspan, [:math:`m`]

    Returns
    -------
    npt.NDArray[np.float_]
        wake vortex separation, [:math:`m`]
    """
    return (np.pi * wingspan) / 4.0


def turbulent_kinetic_energy_dissipation_rate(
    ds_dz: npt.NDArray[np.float_],
    shear_enhancement_factor: npt.NDArray[np.float_] | float = 1.0,
) -> npt.NDArray[np.float_]:
    """
    Calculate the turbulent kinetic energy dissipation rate (epsilon).

    The shear enhancement factor is used to account for any sub-grid scale turbulence.

    Parameters
    ----------
    ds_dz : npt.NDArray[np.float_]
        Difference in wind speed over dz in the atmosphere, [:math:`m s^{-1} / m`]
    shear_enhancement_factor : npt.NDArray[np.float_] | float
        Multiplication factor to enhance the wind shear

    Returns
    -------
    npt.NDArray[np.float_]
        turbulent kinetic energy dissipation rate

    Notes
    -----
    See eq. (37) in :cite:`schumannContrailCirrusPrediction2012`.

    References
    ----------
    - :cite:`schumannContrailCirrusPrediction2012`
    """
    return 0.5 * 0.1**2 * (ds_dz * shear_enhancement_factor) ** 2


def normalized_dissipation_rate(
    epsilon: npt.NDArray[np.float_],
    wingspan: npt.NDArray[np.float_] | float,
    true_airspeed: npt.NDArray[np.float_],
    aircraft_mass: npt.NDArray[np.float_] | float,
    rho_air: npt.NDArray[np.float_] | float,
) -> npt.NDArray[np.float_]:
    """
    Calculate the normalized dissipation rate of the sinking wake vortex.

    Parameters
    ----------
    epsilon: npt.NDArray[np.float_]
        turbulent kinetic energy dissipation rate
    wingspan : npt.NDArray[np.float_] | float
        aircraft wingspan, [:math:`m`]
    true_airspeed : npt.NDArray[np.float_]
        true airspeed for each waypoint, [:math:`m s^{-1}`]
    aircraft_mass : npt.NDArray[np.float_]
        aircraft mass for each waypoint, [:math:`kg`]
    rho_air : npt.NDArray[np.float_]
        density of air for each waypoint, [:math:`kg m^{-3}`]

    Returns
    -------
    npt.NDArray[np.float_]
        Normalized dissipation rate of the sinking wake vortex

    Notes
    -----
    See page 548 of :cite:`schumannContrailCirrusPrediction2012`.

    References
    ----------
    - :cite:`schumannContrailCirrusPrediction2012`
    """
    c = (np.pi / 4) ** (1 / 3) * np.pi**3 / 8  # This is W6 in Schumann's Fortran code
    numer = c * (epsilon * wingspan) ** (1 / 3) * wingspan**2 * rho_air * true_airspeed

    # epsn_st = epsilon star
    epsn_st = numer / (constants.g * aircraft_mass)

    # In a personal correspondence, Schumann gives the precise value
    # of 0.358906526 here
    # In the 2012 paper, Schumann gives 0.36
    # The precise value is likely insignificant because we don't expect epsn_st
    # to be larger than 0.36
    return np.minimum(epsn_st, 0.36)


def initial_contrail_width(
    wingspan: npt.NDArray[np.float_] | float, dz_max: npt.NDArray[np.float_]
) -> npt.NDArray[np.float_]:
    """
    Calculate the initial contrail width.

    Parameters
    ----------
    wingspan : npt.NDArray[np.float_] | float
        aircraft wingspan, [:math:`m`]
    dz_max : npt.NDArray[np.float_]
        Max contrail downward displacement after the wake vortex phase, [:math:`m`]
        Only the size of this array is used; the values are ignored.

    Returns
    -------
    npt.NDArray[np.float_]
        Initial contrail width, [:math:`m`]
    """
    return np.full_like(dz_max, np.pi / 4) * wingspan


def initial_contrail_depth(
    dz_max: npt.NDArray[np.float_], initial_wake_vortex_depth: float | npt.NDArray[np.float_]
) -> npt.NDArray[np.float_]:
    """
    Calculate the initial contrail depth.

    Parameters
    ----------
    dz_max : npt.NDArray[np.float_]
        Max contrail downward displacement after the wake vortex phase, [:math:`m`]
    initial_wake_vortex_depth : float | npt.NDArray[np.float_]
        Initial wake vortex depth scaling factor.
        Denoted `C_D0` in eq (14) in :cite:`schumannContrailCirrusPrediction2012`.

    Returns
    -------
    npt.NDArray[np.float_]
        Initial contrail depth [m]
    """
    return dz_max * initial_wake_vortex_depth



"""ATC Flight Plan Parser."""

import re
from typing import Any


def to_atc_plan(plan: dict[str, Any]) -> str:
    """Write dictionary from :func:`parse_atc_plan` as ATC flight plan string.

    Parameters
    ----------
    plan: dict[str, Any]
        Dictionary representation of ATC flight plan returned from :func:`parse_atc_plan`.

    Returns
    -------
    str
        ATC flight plan string conforming to ICAO Doc 4444-ATM/501

    See Also
    --------
    :func:`parse_atc_plan`
    """
    ret = f'(FPL-{plan["callsign"]}-{plan["flight_rules"]}'
    ret += f'{plan["type_of_flight"]}\n'
    ret += "-"
    if "number_aircraft" in plan and plan["number_aircraft"] <= 10:
        ret += plan["number_aircraft"]
    ret += f'{plan["type_of_aircraft"]}/{plan["wake_category"]}-'
    ret += f'{plan["equipment"]}/{plan["transponder"]}\n'
    ret += f'-{plan["departure_icao"]}{plan["time"]}\n'
    ret += f'-{plan["speed_type"]}{plan["speed"]}{plan["level_type"]}'
    ret += f'{plan["level"]} {plan["route"]}\n'
    if "destination_icao" in plan and "duration" in plan:
        ret += f'-{plan["destination_icao"]}{plan["duration"]}'
    if "alt_icao" in plan:
        ret += f' {plan["alt_icao"]}'
    if "second_alt_icao" in plan:
        ret += f' {plan["second_alt_icao"]}'
    ret += "\n"
    ret += f'-{plan["other_info"]})\n'
    if "supplementary_info" in plan:
        ret += " ".join([f"{i[0]}/{i[1]}" for i in plan["supplementary_info"].items()])

    if ret[-1] == "\n":
        ret = ret[:-1]

    return ret


def parse_atc_plan(atc_plan: str) -> dict[str, str]:
    """Parse an ATC flight plan string into a dictionary.

    The route string is not converted to lat/lon in this process.

    Parameters
    ----------
    atc_plan : str
        An ATC flight plan string conforming to ICAO Doc 4444-ATM/501 (Appendix 2)

    Returns
    -------
    dict[str, str]
        A dictionary consisting of parsed components of the ATC flight plan.
        A full ATC plan will contain the keys:

        - ``callsign``: ICAO flight callsign
        - ``flight_rules``: Flight rules ("I", "V", "Y", "Z")
        - ``type_of_flight``: Type of flight ("S", "N", "G", "M", "X")
        - ``number_aircraft``: The number of aircraft, if more than one
        - ``type_of_aircraft``: ICAO aircraft type
        - ``wake_category``: Wake turbulence category
        - ``equipment``: Radiocommunication, navigation and approach aid equipment and capabilities
        - ``transponder``: Surveillance equipment and capabilities
        - ``departure_icao``: ICAO departure airport
        - ``time``: Estimated off-block (departure) time (UTC)
        - ``speed_type``: Speed units ("K": km / hr, "N": knots)
        - ``speed``: Cruise true airspeed in ``speed_type`` units
        - ``level_type``: Level units ("F", "S", "A", "M")
        - ``level``: Cruise level
        - ``route``: Route string
        - ``destination_icao``: ICAO destination airport
        - ``duration``: The total estimated elapsed time for the flight plan
        - ``alt_icao``: ICAO alternate destination airport
        - ``second_alt_icao``: ICAO second alternate destination airport
        - ``other_info``: Other information
        - ``supplementary_info``: Supplementary information

    References
    ----------
    - https://applications.icao.int/tools/ATMiKIT/story_content/external_files/story_content/external_files/DOC%204444_PANS%20ATM_en.pdf

    See Also
    --------
    :func:`to_atc_plan`
    """  # noqa: E501
    atc_plan = atc_plan.replace("\r", "")
    atc_plan = atc_plan.replace("\n", "")
    atc_plan = atc_plan.upper()
    atc_plan = atc_plan.strip()

    if len(atc_plan) == 0:
        raise ValueError("Empty or invalid flight plan")

    atc_plan = atc_plan.replace("(FPL", "")
    atc_plan = atc_plan.replace(")", "")
    atc_plan = atc_plan.replace("--", "-")

    basic = atc_plan.split("-")

    flightplan: dict[str, Any] = {}

    # Callsign
    if len(basic) > 1:
        flightplan["callsign"] = basic[1]

    # Flight Rules
    if len(basic) > 2:
        flightplan["flight_rules"] = basic[2][0]
        flightplan["type_of_flight"] = basic[2][1]

    # Aircraft
    if len(basic) > 3:
        aircraft = basic[3].split("/")
        matches = re.match("(\d{1})(\S{3,4})", aircraft[0])
        if matches:
            groups = matches.groups()
        else:
            groups = ()

        if matches and len(groups) > 2:
            flightplan["number_aircraft"] = groups[1]
            flightplan["type_of_aircraft"] = groups[2]
        else:
            flightplan["type_of_aircraft"] = aircraft[0]

        if len(aircraft) > 1:
            flightplan["wake_category"] = aircraft[1]

    # Equipment
    if len(basic) > 4:
        equip = basic[4].split("/")
        flightplan["equipment"] = equip[0]
        if len(equip) > 1:
            flightplan["transponder"] = equip[1]

    # Dep. airport info
    if len(basic) > 5:
        matches = re.match("(\D*)(\d*)", basic[5])
        if matches:
            groups = matches.groups()
        else:
            groups = ()

        if len(groups) > 0:
            flightplan["departure_icao"] = groups[0]
        if len(groups) > 1:
            flightplan["time"] = groups[1]

    # Speed and route info
    if len(basic) > 6:
        matches = re.match("(\D*)(\d*)(\D*)(\d*)", basic[6])
        if matches:
            groups = matches.groups()
        else:
            groups = ()

        # match speed and level
        if len(groups) > 0:
            flightplan["speed_type"] = groups[0]
            if len(groups) > 1:
                flightplan["speed"] = groups[1]
            if len(groups) > 2:
                flightplan["level_type"] = groups[2]
            if len(groups) > 3:
                flightplan["level"] = groups[3]

            flightplan["route"] = basic[6][len("".join(groups)) :].strip()
        else:
            flightplan["route"] = basic[6].strip()

    # Dest. airport info
    if len(basic) > 7:
        matches = re.match("(\D{4})(\d{4})", basic[7])
        if matches:
            groups = matches.groups()
        else:
            groups = ()

        if len(groups) > 0:
            flightplan["destination_icao"] = groups[0]
        if len(groups) > 1:
            flightplan["duration"] = groups[1]

        matches = re.match("(\D{4})(\d{4})(\s{1})(\D{4})", basic[7])
        if matches:
            groups = matches.groups()
        else:
            groups = ()

        if len(groups) > 3:
            flightplan["alt_icao"] = groups[3]

        matches = re.match("(\D{4})(\d{4})(\s{1})(\D{4})(\s{1})(\D{4})", basic[7])
        if matches:
            groups = matches.groups()
        else:
            groups = ()

        if len(groups) > 5:
            flightplan["second_alt_icao"] = groups[5]

    # Other info
    if len(basic) > 8:
        flightplan["other_info"] = basic[8]

    # Supl. Info
    if len(basic) > 9:
        sup_match = re.findall("(\D{1}[\/]{1})", basic[9])
        if len(sup_match) > 0:
            suplInfo = {}
            for i in range(len(sup_match) - 1):
                this_key = sup_match[i]
                this_idx = basic[9].find(this_key)

                next_key = sup_match[i + 1]
                next_idx = basic[9].find(next_key)

                val = basic[9][this_idx + 2 : next_idx - 1]
                suplInfo[this_key[0]] = val

            last_key = sup_match[-1]
            last_idx = basic[9].find(last_key)
            suplInfo[last_key[0]] = basic[9][last_idx + 2 :]

            flightplan["supplementary_info"] = suplInfo

    return flightplan



"""Utilites for iterating of sequences."""

from __future__ import annotations

from typing import Any, Iterator


def chunk_list(lst: list, n: int) -> Iterator[list[Any]]:
    """Yield successive n-sized chunks from list."""

    for i in range(0, len(lst), n):
        yield lst[i : i + n]



"""Schmidt-Appleman criteria (SAC)."""

from __future__ import annotations

import dataclasses
from typing import Any, overload

import numpy as np
import scipy.optimize

import pycontrails
from pycontrails.core.flight import Flight
from pycontrails.core.fuel import Fuel, JetA
from pycontrails.core.met import MetDataArray, MetDataset
from pycontrails.core.met_var import AirTemperature, SpecificHumidity
from pycontrails.core.models import Model, ModelParams
from pycontrails.core.vector import GeoVectorDataset
from pycontrails.models.humidity_scaling import HumidityScaling
from pycontrails.physics import constants, thermo
from pycontrails.utils.types import ArrayLike, ArrayScalarLike, apply_nan_mask_to_arraylike

# -----------------
# Models as classes
# -----------------


@dataclasses.dataclass
class SACParams(ModelParams):
    """Parameters for :class:`SAC`."""

    #: Jet engine efficiency, [:math:`0 - 1`]
    engine_efficiency: float = 0.3

    #: Fuel type.
    #: Overridden by Fuel provided on input ``source`` attributes
    fuel: Fuel = dataclasses.field(default_factory=JetA)

    #: Humidity scaling
    humidity_scaling: HumidityScaling | None = None


class SAC(Model):
    """Determine points where Schmidt-Appleman Criteria is satisfied.

    Parameters
    ----------
    met : MetDataset
        Dataset containing "air_temperature", "specific_humidity" variables.
    params : dict[str, Any], optional
        Override :class:`SACParams` with dictionary.
    **params_kwargs
        Override :class:`SACParams` with keyword arguments.
    """

    name = "sac"
    long_name = "Schmidt-Appleman contrail formation criteria"
    met_variables = AirTemperature, SpecificHumidity
    default_params = SACParams

    @overload
    def eval(self, source: Flight, **params: Any) -> Flight:
        ...

    @overload
    def eval(self, source: GeoVectorDataset, **params: Any) -> GeoVectorDataset:
        ...

    @overload
    def eval(self, source: MetDataset | None = None, **params: Any) -> MetDataArray:
        ...

    def eval(
        self, source: GeoVectorDataset | Flight | MetDataset | None = None, **params: Any
    ) -> GeoVectorDataset | Flight | MetDataArray:
        """Evaluate the Schmidt-Appleman criteria along flight trajectory or on meteorology grid.

        .. versionchanged:: 0.27.0

            Humidity scaling now handled automatically. This is controlled by
            model parameter ``humidity_scaling``.

        Parameters
        ----------
        source : GeoVectorDataset | Flight | MetDataset | None, optional
            Input GeoVectorDataset or Flight.
            If None, evaluates at the :attr:`met` grid points.
        **params : Any
            Overwrite model parameters before eval

        Returns
        -------
        GeoVectorDataset | Flight | MetDataArray
            Returns 1 where SAC is satisfied, 0 everywhere else.
            Returns `np.nan` if interpolating outside meteorology grid.

        Raises
        ------
        NotImplementedError
            Raises if input ``source`` is not supported.
        """

        self.update_params(params)
        self.set_source(source)
        self.require_source_type((GeoVectorDataset, MetDataset))

        if isinstance(self.source, GeoVectorDataset):
            self.downselect_met()
            self.source.setdefault("air_pressure", self.source.air_pressure)

        scale_humidity = (self.params["humidity_scaling"] is not None) and (
            "specific_humidity" not in self.source
        )
        self.set_source_met()

        # apply humidity scaling, warn if no scaling is provided for ECMWF data
        if scale_humidity:
            self.params["humidity_scaling"].eval(self.source, copy_source=False)

        # Extract source data
        air_temperature = self.source.data["air_temperature"]
        specific_humidity = self.source.data["specific_humidity"]
        air_pressure = self.source.data["air_pressure"]
        engine_efficiency = self.get_source_param("engine_efficiency")

        # Flight class has fuel attribute, use this instead of params
        if isinstance(self.source, Flight):
            fuel = self.source.fuel
        else:
            fuel = self.get_source_param("fuel")
        assert isinstance(fuel, Fuel), "The fuel attribute must be of type Fuel"

        ei_h2o = fuel.ei_h2o
        q_fuel = fuel.q_fuel

        G = slope_mixing_line(specific_humidity, air_pressure, engine_efficiency, ei_h2o, q_fuel)
        T_sat_liquid_ = T_sat_liquid(G)
        rh_crit_sac = rh_critical_sac(air_temperature, T_sat_liquid_, G)
        rh = thermo.rh(specific_humidity, air_temperature, air_pressure)
        sac_ = sac(rh, rh_crit_sac)

        # Attaching some intermediate artifacts onto the source
        self.source["G"] = G
        self.source["T_sat_liquid"] = T_sat_liquid_
        self.source["rh"] = rh
        self.source["rh_critical_sac"] = rh_crit_sac
        self.source["sac"] = sac_

        # Tag output with additional attrs when source is MetDataset
        if isinstance(self.source, MetDataset):
            attrs: dict[str, Any] = {
                "description": self.long_name,
                "pycontrails_version": pycontrails.__version__,
            }
            if scale_humidity:
                for k, v in self.params["humidity_scaling"].description.items():
                    attrs[f"humidity_scaling_{k}"] = v
            if self.met is not None:
                attrs["met_source"] = self.met.attrs.get("met_source", "unknown")
            if isinstance(engine_efficiency, (int, float)):
                attrs["engine_efficiency"] = engine_efficiency

            self.source["sac"].data.attrs.update(attrs)
            return self.source["sac"]

        # In GeoVectorDataset case, return source
        return self.source


# -------------------
# Models as functions
# -------------------


def slope_mixing_line(
    specific_humidity: ArrayLike,
    air_pressure: ArrayLike,
    engine_efficiency: float | ArrayLike,
    ei_h2o: float,
    q_fuel: float,
) -> ArrayLike:
    r"""Calculate the slope of the mixing line in a temperature-humidity diagram.

    This quantity is often notated with ``G`` in the literature.

    Parameters
    ----------
    specific_humidity : ArrayLike
        A sequence or array of specific humidity values, [:math:`kg_{H_{2}O} \ kg_{air}`]
    air_pressure : ArrayLike
        A sequence or array of atmospheric pressure values, [:math:`Pa`].
    engine_efficiency: float | ArrayLike
        Engine efficiency, [:math:`0 - 1`]
    ei_h2o : float
        Emission index of water vapor, [:math:`kg \ kg^{-1}`]
    q_fuel : float
        Specific combustion heat of fuel combustion, [:math:`J \ kg^{-1} \ K^{-1}`]

    Returns
    -------
    ArrayLike
        Slope of the mixing line in a temperature-humidity diagram, [:math:`Pa \ K^{-1}`]
    """
    c_pm = thermo.c_pm(specific_humidity)  # Often taken as 1004 (= constants.c_pd)
    return (ei_h2o * c_pm * air_pressure) / (constants.epsilon * q_fuel * (1.0 - engine_efficiency))


def T_sat_liquid(G: ArrayLike) -> ArrayLike:
    r"""Calculate temperature at which liquid saturation curve has slope G.

    Parameters
    ----------
    G : ArrayLike
        Slope of the mixing line in a temperature-humidity diagram.

    Returns
    -------
    ArrayLike
        Maximum threshold temperature for 100% relative humidity with respect to liquid,
        [:math:`K`]. This can also be interpreted as the temperature at which the liquid
        saturation curve has slope G.

    References
    ----------
    - :cite:`schumannConditionsContrailFormation1996`

    See Also
    --------
    :func:`T_sat_liquid_high_accuracy`

    Notes
    -----
    Defined (using notation T_LM) in :cite:`schumannConditionsContrailFormation1996`
    in the first full paragraph on page 10 as

      for T = T_LC, the mixing line just touches [is tangent to]
      the saturation curve. See equation (10).

    The formula used here is taken from equation (31).
    """
    # FIXME: Presently, mypy is not aware that numpy ufuncs will return `xr.DataArray``
    # when xr.DataArray is passed in. This will get fixed at some point in the future
    # as `numpy` their typing patterns, after which the "type: ignore" comment can
    # get ripped out.
    # We could explicitly check for `xr.DataArray` then use `xr.apply_ufunc`, but
    # this only renders our code more boilerplate and less performant.
    # This comment is pasted several places in `pycontrails` -- they should all be
    # addressed at the same time.
    log_ = np.log(G - 0.053)
    return -46.46 - constants.absolute_zero + 9.43 * log_ + 0.72 * log_**2  # type: ignore[return-value]  # noqa: E501


def _e_sat_liquid_prime(T: ArrayScalarLike) -> ArrayScalarLike:
    r"""Calculate derivative of :func:`thermo.e_sat_liquid`.

    Parameters
    ----------
    T : ArrayScalarLike
        Temperature, [:math:`K`].

    Returns
    -------
    ArrayScalarLike
        Derivative of :func:`thermo.e_sat_liquid`, [:math:``Pa \ K^{-1}`].
    """
    d_inside = 6096.9385 / (T**2) - 0.02711193 + 1.673952 * 1e-5 * 2 * T + 2.433502 / T
    return thermo.e_sat_liquid(T) * d_inside


def T_sat_liquid_high_accuracy(
    G: ArrayLike,
    maxiter: int = 5,
) -> ArrayLike:
    """Calculate temperature at which liquid saturation curve has slope G.

    The function :func:`T_sat_liquid` gives a first order approximation to equation (10)
    of the Schumann paper referenced below. This function uses Newton's method to
    compute the numeric solution to (10).

    Parameters
    ----------
    G : ArrayLike
        Slope of the mixing line
    maxiter : int, optional
        Passed into :func:`scipy.optimize.newton`. Because ``T_sat_liquid`` is already
        fairly accurate, few iterations are needed for Newton's method to converge.
        By default, 5.

    Returns
    -------
    ArrayLike
        Maximum threshold temperature for 100% relative humidity with respect to liquid,
        [:math:`K`].

    References
    ----------
    - :cite:`schumannConditionsContrailFormation1996`

    See Also
    --------
    :func:`T_sat_liquid_high`
    """
    init_guess = T_sat_liquid(G)

    def func(T: ArrayLike) -> ArrayLike:
        """Equation (10) from Schumann 1996."""
        return _e_sat_liquid_prime(T) - G

    return scipy.optimize.newton(func, init_guess, maxiter=maxiter)


def rh_critical_sac(air_temperature: ArrayLike, T_sat_liquid: ArrayLike, G: ArrayLike) -> ArrayLike:
    r"""Calculate critical relative humidity threshold of contrail formation.

    Parameters
    ----------
    air_temperature : ArrayLike
        A sequence or array of temperature values, [:math:`K`]
    T_sat_liquid : ArrayLike,
        Maximum threshold temperature for 100% relative humidity with respect to liquid, [:math:`K`]
    G : ArrayLike
        Slope of the mixing line in a temperature-humidity diagram.

    Returns
    -------
    ArrayLike
        Critical relative humidity of contrail formation, [:math:`[0 - 1]`]

    References
    ----------
    - :cite:`ponaterContrailsComprehensiveGlobal2002`
    """
    e_sat_T_sat_liquid = thermo.e_sat_liquid(T_sat_liquid)  # always positive
    e_sat_T = thermo.e_sat_liquid(air_temperature)  # always positive

    # Below, `rh_crit` can be negative
    rh_crit = (G * (air_temperature - T_sat_liquid) + e_sat_T_sat_liquid) / e_sat_T

    # Per Ponater, section 2.3:
    # >>> The critical relative humidity `r_contr` can range from 0 to 1
    # The "first order" term `G * (air_temperature - T_sat_liquid)` can be negative.
    # Consequently, we clip rh_crit at 0 and 1.
    # After clipping, when rh_crit = 0 the SAC is guaranteed to be satisfied.

    # Per Ponater, beginning of section 2.3:
    # >>> "No contrails are possible if T > T_contr" <<<
    # In our notation, this inequality is `air_temperature > T_sat_liquid`
    # We set the corresponding rh_crit to infinity, indicating no SAC is possible

    # numpy case
    if isinstance(rh_crit, np.ndarray):
        rh_crit.clip(0.0, 1.0, out=rh_crit)  # clip in place
        rh_crit[air_temperature > T_sat_liquid] = np.inf
        return rh_crit

    # xarray case
    # FIXME: the two cases handle nans differently. Unfortunately, unit tests break
    # if I try to consolidate to a single condition
    rh_crit = rh_crit.clip(0.0, 1.0)
    return rh_crit.where(air_temperature <= T_sat_liquid, np.inf)


def sac(
    rh: ArrayLike,
    rh_crit_sac: ArrayLike,
) -> ArrayLike:
    r"""Points at which the Schmidt-Appleman Criteria is satisfied.

    Parameters of type :class:`ArrayLike` must have compatible shapes.

    Parameters
    ----------
    rh : ArrayLike
        Relative humidity values
    rh_crit_sac: ArrayLike
        Critical relative humidity threshold of contrail formation

    Returns
    -------
    ArrayLike
        SAC state of each point indexed by the :class:`ArrayLike` parameters.
        Returned array has floating ``dtype`` with values

            - 0.0 signifying SAC fails
            - 1.0 signifying SAC holds

        NaN entries of parameters propagate into the returned array.
    """
    nan_mask = np.isnan(rh) | np.isnan(rh_crit_sac)

    dtype = np.result_type(rh, rh_crit_sac)
    sac_ = (rh > rh_crit_sac).astype(dtype)
    return apply_nan_mask_to_arraylike(sac_, nan_mask)


def T_critical_sac(
    T_LM: ArrayLike,
    relative_humidity: ArrayLike,
    G: ArrayLike,
    maxiter: int = 10,
) -> ArrayLike:
    r"""Estimate temperature threshold for persistent contrail formation.

    This quantity is defined as ``T_LC`` in Schumann (see reference below). Equation (11)
    of this paper implicitly defines ``T_LC`` as the solution to the equation
    ::

        T_LC = T_LM - (e_L(T_LM) - rh * e_L(T_LC)) / G

    For relative humidity above 0.999, the corresponding entry from ``T_LM``
    is returned (page 10, top of the right-hand column). Otherwise, the solution
    to the equation above is approximated via Newton's method.

    Parameters
    ----------
    T_LM : ArrayLike
        Output of :func:`T_sat_liquid` calculation.
    relative_humidity : ArrayLike
        Relative humidity values
    G : ArrayLike
        Slope of the mixing line in a temperature-humidity diagram.
    maxiter : int, optional
        Passed into :func:`scipy.optimize.newton`. By default, 10.

    Returns
    -------
    ArrayLike
        Critical temperature threshold values.

    References
    ----------
    - :cite:`schumannConditionsContrailFormation1996`
    """
    # Near U = 1, Newton's method is slow to converge (I believe this is because
    # the function `func` has a double root at T_LM when U = 1, so Newton's method
    # is somewhat degenerate here)
    # But the answer is known in this case. This is discussed at the top of the
    # right hand column on page 10 in Schumann 1996.
    # We only apply Newton's method at points with rh bounded below 1 (scipy will
    # raise an error if Newton's method is not converging well).
    filt = relative_humidity < 0.999
    if not np.any(filt):
        return T_LM

    U_filt = relative_humidity[filt]
    T_LM_filt = T_LM[filt]
    e_L_of_T_LM_filt = thermo.e_sat_liquid(T_LM_filt)
    G_filt = G[filt]

    def func(T: ArrayLike) -> ArrayLike:
        """Equation (11) from Schumann."""
        return T - T_LM_filt + (e_L_of_T_LM_filt - U_filt * thermo.e_sat_liquid(T)) / G_filt

    def fprime(T: ArrayLike) -> ArrayLike:
        return 1 - U_filt * _e_sat_liquid_prime(T) / G_filt

    # This initial guess should be less than T_LM.
    # For relative_humidity away from 1, Newton's method converges quickly, and so
    # any initial guess will work.
    init_guess = T_LM_filt - 1
    newton_approx = scipy.optimize.newton(func, init_guess, fprime=fprime, maxiter=maxiter)

    # For relative_humidity > 0.999, we just use T_LM
    # We copy over the entire array T_LM here instead of using np.empty_like
    # in order to keep typing compatible with xarray types
    out = T_LM.copy()
    out[filt] = newton_approx
    return out



"""Support methods for validation scripts."""

from __future__ import annotations

import numpy as np
import pandas as pd
from numpy.testing import assert_allclose, assert_array_equal

from pycontrails import Flight
from pycontrails.models import Cocip
from pycontrails.physics import units

FLIGHT_TOLERANCE = 1e-4
CONTRAIL_TOLERANCE = 1e-3


def create_flight(flight_id: str, df_flight: pd.DataFrame) -> Flight:
    """
    Function to create a Flight class from a
    group of rows with the same Flight ID
    """

    # constant properties along the length of the flight
    attrs = {
        "flight_id": flight_id,
        # get first val of atyp col as aircraft type
        "aircraft_type": df_flight["ATYP"].values[0],
    }

    # # DISCUSSION: We could use the Aircraft class here
    # aircraft = Aircraft(aircraft_type=df_flight["ATYP"].values[0])

    # convert UTC timestamp to np.datetime64
    df_flight["time"] = pd.to_datetime(df_flight["date_time"])

    # get altitude in m
    df_flight["altitude"] = df_flight["alt_ft"] * 0.3048

    # rename a few columns for compatibility with `Flight` requirements
    df_flight = df_flight.rename(columns={"lon_deg": "longitude", "lat_deg": "latitude"})

    # clean up a few columns before building Flight class
    df_flight = df_flight.drop(columns=["Flight ID", "date_time", "ATYP", "alt_ft"])

    # create a Flight class
    return Flight(data=df_flight, attrs=attrs)


def validate_output(
    cocip: Cocip,
    pycocip_flight: pd.DataFrame | None,
    pycocip_flight_stats: pd.Series,
    pycocip_contrail: pd.DataFrame,
) -> None:
    # ---
    # Flight waypoint validation
    # ---
    flight = cocip.flight

    # Test flight waypoints
    if pycocip_flight is not None:
        assert_array_equal(flight["time"], pd.to_datetime(pycocip_flight["time"]), err_msg="time")
        assert_allclose(flight["longitude"], pycocip_flight["lon"], err_msg="longitude")
        assert_allclose(flight["latitude"], pycocip_flight["lat"], err_msg="latitude")
        assert_allclose(
            units.m_to_ft(flight.altitude), pycocip_flight["alt_ft"], err_msg="altitude"
        )
        assert_allclose(flight.air_pressure, pycocip_flight["pressure_pa"], err_msg="level")
        assert_allclose(
            flight["air_temperature"], pycocip_flight["temperature"], err_msg="air_temperature"
        )
        assert_allclose(flight["u_wind"], pycocip_flight["winds_u"], err_msg="u_wind")
        assert_allclose(flight["v_wind"], pycocip_flight["winds_v"], err_msg="v_wind")
        assert_allclose(
            flight["segment_length"], pycocip_flight["seg_length"], err_msg="segment_length"
        )
        assert_allclose(flight["sin_a"], pycocip_flight["sin_a"], err_msg="sin_a")
        assert_allclose(flight["cos_a"], pycocip_flight["cos_a"], err_msg="cos_a")
        assert_allclose(
            flight["true_airspeed"],
            pycocip_flight["tas_ms"],
            rtol=FLIGHT_TOLERANCE,
            err_msg="true_airspeed",
        )
        assert_allclose(
            flight["aircraft_mass"],
            pycocip_flight["amass"],
            rtol=FLIGHT_TOLERANCE,
            err_msg="aircraft_mass",
        )
        assert_allclose(
            flight["fuel_flow"],
            pycocip_flight["fuel_flow"],
            rtol=FLIGHT_TOLERANCE,
            err_msg="fuel_flow",
        )
        assert_allclose(
            flight["engine_efficiency"],
            pycocip_flight["ope"],
            rtol=FLIGHT_TOLERANCE,
            err_msg="engine_efficiency",
        )
        assert_allclose(
            flight["thrust"],
            pycocip_flight["thrust_setting"],
            rtol=FLIGHT_TOLERANCE,
            err_msg="thrust",
        )
        assert_allclose(
            pycocip_flight["bc_ei_n"],
            flight["nvpm_ei_n"],
            rtol=FLIGHT_TOLERANCE,
            err_msg="nvpm_ei_n",
        )
        assert np.all((flight["sac"] == 1) == pycocip_flight["has_initial_contrails"])
        assert_allclose(
            flight["T_critical_sac"],
            pycocip_flight["temp_lc"],
            rtol=FLIGHT_TOLERANCE,
            err_msg="T_critical_sac",
        )
        assert_allclose(
            flight["rhi"],
            pycocip_flight["initial_rhi"],
            rtol=CONTRAIL_TOLERANCE,
            err_msg="rhi",
        )
        # assert_allclose(flight.data["nox_ei"], pycocip_flight["nox_ei"])

    # Test flight stats
    assert flight.attrs["aircraft_type"] == pycocip_flight_stats["ATYP"]
    assert flight.attrs["fuel_data_source"] == pycocip_flight_stats["Fuel methodology"]
    assert flight.attrs["engine_name"] == pycocip_flight_stats["Engine name"]
    assert flight.attrs["bc_data_source"] == pycocip_flight_stats["BC methodology"]
    assert flight.attrs["aircraft_type_bada"] == pycocip_flight_stats["ATYP assumed"]

    assert flight["time"][0] == pd.to_datetime(pycocip_flight_stats["First wypt"])
    assert flight["time"][-1] == pd.to_datetime(pycocip_flight_stats["Last wypt"])
    assert_allclose(
        flight.length / 1000.0,
        pycocip_flight_stats["Total flight dist (km)"],
        rtol=FLIGHT_TOLERANCE,
        err_msg="length",
    )
    assert_allclose(
        flight.attrs["total_fuel_burn"],
        pycocip_flight_stats["Total fuel burn (kg)"],
        rtol=FLIGHT_TOLERANCE,
        err_msg="total_fuel_burn",
    )
    assert_allclose(
        np.nanmean(flight["nvpm_ei_n"]),
        pycocip_flight_stats["BC EI_n mean (kg-1)"],
        rtol=FLIGHT_TOLERANCE,
        err_msg="nvpm_ei_n mean",
    )
    assert_allclose(
        np.nanmean(flight["aircraft_mass"]),
        pycocip_flight_stats["Aircraft mass mean (kg)"],
        rtol=FLIGHT_TOLERANCE,
        err_msg="aircraft_mass mean",
    )
    assert_allclose(
        np.nanmean(flight["engine_efficiency"]),
        pycocip_flight_stats["OPE mean"],
        rtol=FLIGHT_TOLERANCE,
        err_msg="engine_efficiency mean",
    )

    # return if no contrail
    if not len(pycocip_contrail):
        assert cocip.contrail is None or not len(cocip.contrail)
        return

    # ---
    # Contrail waypoint validation
    # ---
    contrail = cocip.contrail
    assert contrail is not None

    # waypoints
    original_waypoints = contrail["waypoint"].apply(lambda w: flight["wypt_id"][w])
    assert_allclose(original_waypoints, pycocip_contrail["wypt_uid"], err_msg="original_waypoints")

    # location
    assert_allclose(
        contrail["latitude"], pycocip_contrail["lat"], rtol=CONTRAIL_TOLERANCE, err_msg="latitude"
    )
    assert_allclose(
        contrail["longitude"], pycocip_contrail["lon"], rtol=CONTRAIL_TOLERANCE, err_msg="longitude"
    )
    assert_allclose(
        units.m_to_ft(contrail["altitude"]),
        pycocip_contrail["alt_ft"].values,
        rtol=CONTRAIL_TOLERANCE,
        err_msg="altitude",
    )

    # continuity
    assert_array_equal(contrail["continuous"], pycocip_contrail["continuous"], err_msg="continuous")
    continuous = contrail["continuous"].values
    assert np.all(contrail[~continuous]["ef"] == 0)

    # downselect contrail for only continuous points
    # pycocip has incorrect values in these locations
    contrail = contrail[continuous]
    pycocip_contrail = pycocip_contrail[
        continuous
    ]  # using `continuous` here since we've already asserted that pycontrails/pycocip agree here

    # geometry
    assert_allclose(
        contrail["segment_length"],
        pycocip_contrail["seg_length"],
        rtol=CONTRAIL_TOLERANCE,
        err_msg="segment_length",
    )
    assert_allclose(
        contrail["sin_a"], pycocip_contrail["sin_a"], rtol=CONTRAIL_TOLERANCE, err_msg="sin_a"
    )
    assert_allclose(
        contrail["cos_a"], pycocip_contrail["cos_a"], rtol=CONTRAIL_TOLERANCE, err_msg="cos_a"
    )

    # met
    assert_allclose(
        contrail["air_temperature"],
        pycocip_contrail["temp"],
        rtol=CONTRAIL_TOLERANCE,
        err_msg="air_temperature",
    )
    assert_allclose(
        contrail["u_wind"], pycocip_contrail["winds_u"], rtol=CONTRAIL_TOLERANCE, err_msg="u_wind"
    )
    assert_allclose(
        contrail["v_wind"], pycocip_contrail["winds_v"], rtol=CONTRAIL_TOLERANCE, err_msg="v_wind"
    )
    assert_allclose(
        contrail["vertical_velocity"],
        pycocip_contrail["omega"],
        rtol=CONTRAIL_TOLERANCE,
        err_msg="vertical_velocity",
    )
    assert_allclose(
        contrail["specific_humidity"],
        pycocip_contrail["q"],
        rtol=CONTRAIL_TOLERANCE,
        err_msg="specific_humidity",
    )
    assert_allclose(
        contrail["rhi"], pycocip_contrail["rhi"], rtol=CONTRAIL_TOLERANCE, err_msg="rhi"
    )
    assert_allclose(
        contrail["dt_dz"], pycocip_contrail["dt_dz"], rtol=CONTRAIL_TOLERANCE, err_msg="dt_dz"
    )
    assert_allclose(
        contrail["ds_dz"], pycocip_contrail["ds_dz"], rtol=CONTRAIL_TOLERANCE, err_msg="ds_dz"
    )
    assert_allclose(
        contrail["dsn_dz"], pycocip_contrail["dsn_dz"], rtol=CONTRAIL_TOLERANCE, err_msg="dsn_dz"
    )
    assert_allclose(
        contrail["tau_cirrus"],
        pycocip_contrail["tau_cirrus"],
        rtol=CONTRAIL_TOLERANCE,
        err_msg="tau_cirrus",
    )

    # contrail properties
    assert_allclose(
        contrail["iwc"], pycocip_contrail["iwc"], rtol=CONTRAIL_TOLERANCE, err_msg="iwc"
    )
    assert_allclose(
        contrail["width"], pycocip_contrail["width"], rtol=CONTRAIL_TOLERANCE, err_msg="width"
    )
    assert_allclose(
        contrail["depth"], pycocip_contrail["depth"], rtol=CONTRAIL_TOLERANCE, err_msg="depth"
    )
    assert_allclose(
        contrail["plume_mass_per_m"],
        pycocip_contrail["plume_mass_per_m"],
        rtol=CONTRAIL_TOLERANCE,
        err_msg="plume_mass_per_m",
    )
    assert_allclose(
        contrail["r_ice_vol"],
        pycocip_contrail["r_ice_vol"],
        rtol=CONTRAIL_TOLERANCE,
        err_msg="r_ice_vol",
    )
    assert_allclose(
        contrail["tau_contrail"],
        pycocip_contrail["tau_contrail"],
        rtol=CONTRAIL_TOLERANCE,
        err_msg="tau_contrail",
    )
    assert_allclose(
        contrail["n_ice_per_m"],
        pycocip_contrail["n_ice_per_m"],
        rtol=CONTRAIL_TOLERANCE,
        err_msg="n_ice_per_m",
    )
    assert_allclose(
        contrail["n_ice_per_vol"],
        pycocip_contrail["n_ice_per_vol"],
        rtol=CONTRAIL_TOLERANCE,
        err_msg="n_ice_per_vol",
    )
    assert_allclose(
        contrail["n_ice_per_kg_air"],
        pycocip_contrail["n_ice_per_kg_air"],
        rtol=CONTRAIL_TOLERANCE,
        err_msg="n_ice_per_kg_air",
    )
    assert_allclose(
        contrail["age"] / np.timedelta64(1, "h"),
        pycocip_contrail["age_hrs"],
        rtol=CONTRAIL_TOLERANCE,
        err_msg="age",
    )
    assert_allclose(
        contrail["age_hours"],
        pycocip_contrail["age_hrs"],
        rtol=CONTRAIL_TOLERANCE,
        err_msg="age_hours",
    )

    # radiative properties
    assert_allclose(
        contrail["sdr"], pycocip_contrail["sdr"], rtol=CONTRAIL_TOLERANCE, err_msg="sdr"
    )
    assert_allclose(
        contrail["rsr"], pycocip_contrail["rsr"], rtol=CONTRAIL_TOLERANCE, err_msg="rsr"
    )
    assert_allclose(
        contrail["olr"], pycocip_contrail["olr"], rtol=CONTRAIL_TOLERANCE, err_msg="olr"
    )
    assert_allclose(
        contrail["rf_sw"], pycocip_contrail["rf_sw"], rtol=CONTRAIL_TOLERANCE, err_msg="rf_sw"
    )
    assert_allclose(
        contrail["rf_lw"], pycocip_contrail["rf_lw"], rtol=CONTRAIL_TOLERANCE, err_msg="rf_lw"
    )
    assert_allclose(
        contrail["rf_net"], pycocip_contrail["rf_net"], rtol=CONTRAIL_TOLERANCE, err_msg="rf_net"
    )

    # energy forcing
    assert_allclose(contrail["ef"], pycocip_contrail["ef"], rtol=CONTRAIL_TOLERANCE, err_msg="ef")
